/* $Id: sfskeyupdate.C,v 1.34 2002/08/21 13:54:45 max Exp $ */

/*
 *
 * Copyright (C) 1999 David Mazieres (dm@uun.org)
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2, or (at
 * your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 * USA
 *
 */


#include "sfskey.h"
#include "srp.h"

struct updatecmd {
  bool opt_nosrp;
  vec<str> authservers;
  str oldkeyname;
  str newkeyname;
  ptr<sfskey> oldkey;
  ptr<sfskey> newkey;
  sfsauth_srpparmsres srpparms;
  u_int npending;
  bool errors;

  struct ucstate {
    ucstate (str p, ptr<sfscon> s) : 
      path (p == "-" ? str ("localhost") : p), sc (s), 
      c (aclnt::alloc (sc->x, sfsauth_prog_2)) {}
    const str path;
    const ptr<sfscon> sc;
    const ref<aclnt> c;
    str realm;
    sfsauth2_query_res aqr;
    sfsauth2_update_res aur;
    sfs_loginres lres;
    AUTH *auth;
  };

  updatecmd () : opt_nosrp (false), errors (false) { 
    srpparms.set_status (SFSAUTH_FAILED); 
  }

  void setsrp () {
    if (opt_nosrp) {
      srpparms.set_status (SFSAUTH_FAILED);
      return;
    }
    if (srpparms.status != SFSAUTH_OK && !getsrpparms (NULL)) 
      fatal << "Cannot procede without valid SRP parameters.\n";
  }

  void gotcertinfo (ucstate *st, clnt_stat err) { 
    sfs_hostinfo_upgrade (st->sc->servinfo);
    str hostname = sfs_get_hostinfo (st->sc->servinfo).hostname;
    sfsauth_certinfores certres = *(st->aqr.certinfo);
    if (err) {
      warnx << hostname << "authserv: certinfo: " << err << "\n";
      done (false, st);
      return ;
    }
    if (certres.name.len () <= 0) {
      warnx << hostname << "could not get server's name/realm";
      done (false, st);
      return ;
    }
    if (certres.info.status == SFSAUTH_CERT_REALM)
      warnx << "Authserver is in realm " << certres.name << "\n";
    st->realm = certres.name;

    dologin (st);
  }

  void getcertinfo (ucstate *st) {
    assert (srpparms.status == SFSAUTH_OK);

    sfsauth2_query_arg aqa;
    aqa.type = SFSAUTH_CERTINFO;
    aqa.key.set_type (SFSAUTH_DBKEY_NULL);
    st->c->call (SFSAUTH2_QUERY, &aqa, &st->aqr,
		    wrap (this, &updatecmd::gotcertinfo, st));
  }

  void done (bool ok, ucstate *s) {
    if (s)
      delete s;
    if (!ok)
      errors = true;
    if (!--npending)
      exit (errors);
  }

  bool setuparg (ucstate *st, sfsauth2_update_arg *aua) {
    assert (sfs_hostinfo_assert (st->sc->servinfo));
    str hostname = sfs_get_hostinfo (st->sc->servinfo).hostname;
    aua->req.type = SFS_UPDATEREQ;
    aua->req.authid = st->sc->authid;
    aua->req.rec.set_type (SFSAUTH_USER);
    sfsauth_userinfo uinfo = *(st->aqr.userinfo);
    if (!newkey->key->export_pubkey (&uinfo.pubkey)) 
      fatal << "Cannot set new public key\n";
    
    if (srpparms.status == SFSAUTH_OK) {
      srp_client srpc;
      uinfo.pwauth = srpc.create (srpparms.parms->N, 
				  srpparms.parms->g,
				  newkey->pwd,
				  st->realm,
				  newkey->cost, 0);
      if (!uinfo.pwauth) {
	warn << hostname << ": could not create SRP info\n";
	return false;
      }
      str epk;
      if (!newkey->key->export_privkey (&epk, &srpc.eksb)) {
	warn << hostname << ": could not encrypt private key\n";
	return false;
      }
      uinfo.privkey = epk;
    } else {
      uinfo.pwauth = "";
      uinfo.privkey = "";
    }
    uinfo.vers ++;
    *(aua->req.rec.userinfo) = uinfo;
    bzero (&uinfo, sizeof (uinfo));
    str rawmsg (xdr2str (aua->req, true));
    if (!rawmsg) {
      warn << hostname << ": could not marshal update arg\n";
      return false;
    }
    aua->newsig.alloc ();
    aua->authsig.alloc ();
    if (!oldkey->key->sign (aua->authsig, rawmsg) || 
	!newkey->key->sign (aua->newsig, rawmsg)) {
      warn << hostname << ": could not sign update information\n";
      return false;
    }
    return true;
  }

  void start () { sfsfetchkey (newkeyname, wrap (this, 
						 &updatecmd::getnewkey)); }

  void getnewkey (ptr<sfskey> sk, str err, ptr<sfscon> sc) {
    if (!sk)
      fatal << err << "\n";
    newkey = sk;
    if (!opt_nosrp && !newkey->pwd)
      fatal << "New key must have passphrase when using SRP\n";
    if (!opt_nosrp && srpparms.status != SFSAUTH_OK && sc 
	&& sc->hostid_valid) {
      getsrpparms (sc);
    }
    if (oldkeyname == newkeyname)
      getoldkey (false, sk, err, sc);
    else
      sfsfetchkey (oldkeyname, 
		   wrap (this, &updatecmd::getoldkey, true));
  }

  bool getsrpparms (ptr<sfscon> sc) {
    ptr<aclnt> c;
    if (sc)
      c = aclnt::alloc (sc->x, sfsauth_prog_2);
    bigint N, g;
    if (!get_srp_params (c, &N, &g)) 
      return false;
    if (!srp_base::checkparam (N, g)) {
      warnx ("invalid SRP parameters retreived\n");
      return false;
    }
    srpparms.set_status (SFSAUTH_OK);
    srpparms.parms->N = N;
    srpparms.parms->g = g;
    return true;
  }

  void getoldkey (bool retry, ptr<sfskey> sk, str err, ptr<sfscon> sc) {
    if (!sk)
      fatal << err << "\n";
    oldkey = sk;
    if (!opt_nosrp && srpparms.status != SFSAUTH_OK 
	&& sc && sc->hostid_valid && retry) {
      getsrpparms (sc);
    }

    setsrp ();
    if (authservers.empty ()) {
      if (!sc)
	fatal ("must specify authserver -a if SRP fails on old key\n");
      npending = 1;
      getcon (sc->path, sc, NULL);
    }
    else {
      npending = authservers.size ();
      for (const str *sp = authservers.base (); 
	   sp < authservers.lim (); sp++) {
	bool check_hostid = !(*sp == "-");
	sfs_connect_path (*sp, SFS_AUTHSERV,
			  wrap (this, &updatecmd::getcon, *sp),
			  true, check_hostid);
      }
    }
  }

  void getcon (str path, ptr<sfscon> sc, str err) {
    if (!sc) {
      if (path == "-") warnx << "localhost";
      else warnx << path;
      warnx << ": FAILED (" << err << ")\n";
      done (false, NULL);
      return;
    }
    ucstate *st = New ucstate (path, sc);
    if (srpparms.status == SFSAUTH_OK)
      getcertinfo (st);
    else
      dologin (st);
  }

  void dologin (ucstate *st) { 
    ref<aclnt> c (aclnt::alloc (st->sc->x, sfs_program_1));
    sfs_loginarg larg;
    sfs_authreq2 req;
    sfs_autharg2 aarg (SFS_AUTHREQ2);
    req.type = SFS_SIGNED_AUTHREQ;
    req.authid = st->sc->authid;
    req.seqno = 0;
    aarg.sigauth->req.type = SFS_SIGNED_AUTHREQ;
    aarg.sigauth->req.authid = st->sc->authid;
    aarg.sigauth->req.seqno = 0;
    if (!oldkey->key->export_pubkey (&aarg.sigauth->key)) {
      warnx << st->path << ": FAILED (could not export old public key "
	    << "in login)\n";
      done (false, st);
      return;
    }

    str msg;
    if (!(msg = xdr2str (aarg.sigauth->req)) || 
	!oldkey->key->sign (&aarg.sigauth->sig, msg)) {
      warnx << st->path << ": FAILED (could not sign authreq)\n";
      done (false, st);
      return ;
    }
    if (!xdr2bytes (larg.certificate, aarg)) {
      warnx << st->path << ": FAILED (could not marshal login certificate)\n";
      done (false, st);
      return ;
    }
    larg.seqno = 0;

    c->call (SFSPROC_LOGIN, &larg, &st->lres, 
	     wrap (this, &updatecmd::doquery, st));
  }

  void doquery (ucstate *st, clnt_stat err) {
    if (err) {
      warnx << st->path << ": FAILED (" << err << ")\n";
      done (false, st);
      return;
    }
    if (st->lres.status != SFSLOGIN_OK || !*st->lres.authno) {
      warnx << st->path << ": FAILED (login / authentication failed)\n";
      done (false, st);
      return;
    }
    st->auth = authuint_create (*st->lres.authno);
    sfsauth2_query_arg aqa;
    aqa.type = SFSAUTH_USER;
    aqa.key.set_type (SFSAUTH_DBKEY_PUBKEY);
    if (!oldkey->key->export_pubkey (aqa.key.key)) {
      warnx << st->path << ": FAILED (could not export old public key)\n";
      done (false, st);
      return;
    }
    st->c->call (SFSAUTH2_QUERY, &aqa, &st->aqr,
	     wrap (this, &updatecmd::doupdate, st));
  }

  void doupdate (ucstate *st, clnt_stat err) {
    if (err) {
      warnx << st->path << ": FAILED (" << err << ")\n";
      done (false, st);
      return;
    }
    if (st->aqr.type == SFSAUTH_ERROR) {
      warnx << st->path << ": FAILED (" << *(st->aqr.errmsg) << ")\n";
      done (false, st);
      return;
    }
    if (st->aqr.type != SFSAUTH_USER || st->aqr.userinfo->vers == 0) {
      warnx << st->path << ": FAILED (oldkey does not exist on server)\n";
      done (false, st);
      return;
    }
    sfsauth2_update_arg aua;
    if (!setuparg (st, &aua)) {
      warnx << st->path << ": FAILED (could not set up args)\n";
      done (false, st);
      return;
    }
    st->c->call (SFSAUTH2_UPDATE, &aua, &st->aur,
	     wrap (this, &updatecmd::getres, st), st->auth);
  }

  void getres (ucstate *st, clnt_stat err) {
    if (err) {
      warnx << st->path << ": FAILED (" << err << ")\n";
      done (false, st);
    } else if (!st->aur.ok) {
      warnx << st->path << ": FAILED (" << *(st->aur.errmsg) << ")\n";
      done (false, st);
    } else {
      warnx << st->path << ": updated\n";
      done (true, st);
    }
  }
};

void
sfskey_update (int argc, char **argv)
{
  updatecmd *uc = New updatecmd ();

  int ch;
  while ((ch = getopt (argc, argv, "Ss:a:")) != -1)
    switch (ch) {
    case 'S':
      uc->opt_nosrp = true;
      if (uc->srpparms.status == SFSAUTH_OK)
	usage ();
      break;
    case 's':
      {
	if (uc->opt_nosrp)
	  usage ();
	if (str parms = file2str (optarg)) {
	  uc->srpparms.set_status (SFSAUTH_OK);
	  if (!import_srp_params (parms, &uc->srpparms.parms->N, 
				  &uc->srpparms.parms->g))
	    fatal ("%s: cannot parse srp parameters\n", optarg);
	}
	else
	  fatal ("%s: %m\n", optarg);
	break;
      }
    case 'a':
      uc->authservers.push_back (optarg);
      break;
    default:
      usage ();
      break;
    }

  if (optind + 2 < argc || optind >= argc)
    usage ();
  uc->oldkeyname = argv[optind++];
  if (optind < argc)
    uc->newkeyname = argv[optind];
  else
    uc->newkeyname = defkey ();

  if (uc->authservers.empty () && !iskeyremote (uc->oldkeyname))
    fatal ("must specify authserver -a if old key not via SRP\n");

  uc->start ();
}
