/* $Id: authclnt.C,v 1.19 2002/08/21 13:54:47 max Exp $ */

/*
 *
 * Copyright (C) 2002 David Mazieres (dm@uun.org)
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2, or (at
 * your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 * USA
 *
 */

#include "sfscrypt.h"
#include "sfsauthd.h"
#include "rxx.h"

#if HAVE_GETSPNAM
#include <shadow.h>
#endif /* HAVE_GETSPNAM */

extern "C" char *crypt (const char *, const char *);

static str
timestr ()
{
  char buf[80];
  int n = strftime (buf, sizeof (buf), "%a, %d %b %Y %H:%M:%S %z",
		    localtime (&timenow));
  assert (implicit_cast<size_t> (n) < sizeof (buf));
  return buf;
}

authclnt::authclnt (ref<axprt_crypt> x, const authunix_parms *aup)
  : sfsserv (x),
    authsrv (asrv::alloc (x, sfsauth_prog_2,
			  wrap (this, &authclnt::dispatch))),
    sfsauth_login_srp (aup && !aup->aup_uid)
{
  if (aup) {
    uid.alloc ();
    *uid = aup->aup_uid;
  }
}

ptr<sfspriv>
authclnt::doconnect (const sfs_connectarg *ci,
		     sfs_servinfo *si)
{
  *si = myservinfo;
  return myprivkey;
}

bool
authclnt::setuser (sfsauth2_loginres *resp, const sfsauth_dbrec &ae,
		   const dbfile *dbp)
{
  assert (ae.type == SFSAUTH_USER);

  resp->set_status (SFSLOGIN_OK);
  resp->resok->creds.setsize (1);
  resp->resok->creds[0].set_type (SFS_UNIXCRED);

  str name;
  if (dbp->prefix)
    name = dbp->prefix << "/" << ae.userinfo->name;
  else
    name = ae.userinfo->name;

  resp->resok->creds[0].unixcred->uid = dbp->uidmap->map (ae.userinfo->id);
  if (resp->resok->creds[0].unixcred->uid == badid) {
    resp->set_status (SFSLOGIN_BAD);
    *resp->errmsg = "uid out of range";
    return false;
  }

  resp->resok->creds[0].unixcred->username = name;
  resp->resok->creds[0].unixcred->homedir = "/dev/null";
  resp->resok->creds[0].unixcred->shell = "/dev/null";
  if (dbp->allow_unix_pwd && name.len () == strlen (name)) {
    if (struct passwd *pw = getpwnam (ae.userinfo->name)) {
      resp->resok->creds[0].unixcred->homedir = pw->pw_dir;
      resp->resok->creds[0].unixcred->shell = pw->pw_shell;
    }
  }
  else if (dbp->allow_userdir_shell) {
    // Allow remote users without a local account to specify a 
    // shell and homedir
    resp->resok->creds[0].unixcred->homedir = USERDIR_HOMEDIR;
    resp->resok->creds[0].unixcred->shell = USERDIR_SHELL;
  }

  resp->resok->creds[0].unixcred->gid = dbp->gidmap->map (ae.userinfo->gid);

  vec<u_int32_t> groups;
  findgroups (&groups, name);
  resp->resok->creds[0].unixcred->groups.setsize (groups.size ());
  memcpy (resp->resok->creds[0].unixcred->groups.base (),
	  groups.base (), groups.size () * sizeof (groups[0]));

  return true;
}

void
authclnt::findgroups (vec<u_int32_t> *groups, str name)
{
  groups->clear ();
  for (dbfile *dbp = dbfiles.base (); dbp < dbfiles.lim (); dbp++) {
    str suffix;
    if (!dbp->prefix)
      suffix = name;
    else if (dbp->prefix.len () + 1 < name.len ()
	     && name[dbp->prefix.len ()] == '/'
	     && !memcmp (dbp->prefix.cstr (), name.cstr (),
			 dbp->prefix.len ()))
      suffix = substr (name, dbp->prefix.len () + 1);
    else
      continue;
    ptr<authcursor> ac = dbp->db->open ();
    vec<u_int32_t> gv;
    ac->getgroups (&gv, suffix);
    while (!gv.empty ()) {
      u_int32_t gid = dbp->gidmap->map (gv.pop_front ());
      if (gid != badid)
	groups->push_back (gid);
    }
  }
}


bool
authclnt::checkreq (sfsauth2_loginres *resp, const sfsauth2_loginarg *lap,
		    const sfs_authreq2 *reqp)
{
  if (reqp->type != SFS_SIGNED_AUTHREQ) {
    resp->set_status (SFSLOGIN_BAD);
    *resp->errmsg = strbuf ("bad sfs_authreq2 type %d\n", reqp->type);
  }
  else if (reqp->authid != lap->authid) {
    resp->set_status (SFSLOGIN_BAD);
    *resp->errmsg = "authid mismatch\n";
  }
  else if (reqp->seqno != lap->arg.seqno) {
    resp->set_status (SFSLOGIN_BAD);
    *resp->errmsg = "sequence number mismatch\n";
  }
  else
    return true;
  return false;
}

#if 0
ptr<aclnt>
authclnt::getauthclnt ()
{
  /* XXX - Bit of a kludge; should really just override sfs_login */
  static ptr<aclnt> c;
  static ptr<authclnt> ac;
  if (c)
    return c;
  int fds[2];
  if (socketpair (AF_UNIX, SOCK_STREAM, 0, fds) < 0)
    fatal ("socketpair: %m\n");
  c = aclnt::alloc (axprt_stream::alloc (fds[0]), sfsauth_prog_2);
  ac = New refcounted<authclnt> (axprt_crypt::alloc (fds[1]),
				 (authunix_parms *) NULL, true);
  return c;
}
#endif

void
authclnt::dispatch (svccb *sbp)
{
  if (!sbp) {
    delete this;
    return;
  }
  switch (sbp->proc ()) {
  case SFSAUTH2_NULL:
    sbp->reply (NULL);
    break;
  case SFSAUTH2_LOGIN:
    {
      sfsauth2_loginres res;
      sfsauth_login (&res, sbp->template getarg<sfsauth2_loginarg> (),
		     sfsauth_login_srp);
      sbp->replyref (res);
      break;
    }
  case SFSAUTH2_QUERY:
    sfsauth_query (sbp);
    break;
  case SFSAUTH2_UPDATE:
    sfsauth_update (sbp);
    break;
  default:
    sbp->reject (PROC_UNAVAIL);
    break;
  }
}

void
authclnt::sfsauth_update (svccb *sbp)
{
  const sfsauth_cred *cp;
  if (sbp->getaui () >= credtab.size ()
      || (cp = &credtab[sbp->getaui ()])->type != SFS_UNIXCRED) {
    sbp->reject (AUTH_REJECTEDCRED);
    return;
  }

  sfsauth2_update_res res (false);
  sfsauth_dbkey kname (SFSAUTH_DBKEY_NAME);
  *kname.name = cp->unixcred->username;

  sfsauth_dbrec cdbr;
  dbfile *cdbp;
  if (!get_user_cursor (&cdbp, NULL, &cdbr, kname)
      || cp->unixcred->username != cdbr.userinfo->name
      || cp->unixcred->uid != cdbr.userinfo->id) {
    *res.errmsg = "could not load credential db record";
    sbp->replyref (res);
    return;
  }

  sfsauth2_update_arg *argp = sbp->template getarg<sfsauth2_update_arg> ();
  if (argp->req.type != SFS_UPDATEREQ
      || (argp->req.rec.type != SFSAUTH_USER
	  && argp->req.rec.type != SFSAUTH_GROUP)) {
    *res.errmsg = "invalid request";
    sbp->replyref (res);
    return;
  }

  static rxx adminrx ("(\\A|,)admin(\\z|,)");
  bool admin = cdbp->allow_admin && adminrx.match (cdbr.userinfo->privs);

  str reqxdr = xdr2str (argp->req);
  if (argp->newsig) {
    if (!sfscrpt.verify (argp->req.rec.userinfo->pubkey, *(argp->newsig), 
			 reqxdr)) {
      *res.errmsg = "Signature verification failed for newsig.";
      sbp->replyref (res);
      return;
    }
  }
  if (argp->authsig) {
    if (!sfscrpt.verify (cdbr.userinfo->pubkey, *(argp->authsig), reqxdr)) {
      *res.errmsg = "Signature verification failed for authsig.";
      sbp->replyref (res);
      return;
    }
  }
  else if (!cdbp->allow_unix_pwd) {
    *res.errmsg = "digital signature required";
    sbp->replyref (res);
    return;
  }
  else
    admin = false;

  dbfile *udbp;
  ptr<authcursor> uac;
  sfsauth_dbrec udbr;
  if (argp->req.rec.type == SFSAUTH_USER) {
    *kname.name = argp->req.rec.userinfo->name;
    if (!get_user_cursor (&udbp, &uac, &udbr, kname, true)) {
      *res.errmsg = "could not find or update user's record";
      sbp->replyref (res);
      return;
    }
    if (!admin && (udbr.userinfo->name != cdbr.userinfo->name
		   || udbr.userinfo->id != cdbr.userinfo->id)) {
      /* XXX - ignoring owner field for now */
      *res.errmsg = "access denied";
      sbp->replyref (res);
      return;
    }
    if (argp->req.rec.userinfo->vers != udbr.userinfo->vers + 1) {
      *res.errmsg = "version mismatch";
      sbp->replyref (res);
      return;
    }

    uac->ae.userinfo->vers = argp->req.rec.userinfo->vers;
    uac->ae.userinfo->pubkey = argp->req.rec.userinfo->pubkey;
    uac->ae.userinfo->pwauth = argp->req.rec.userinfo->pwauth;
    uac->ae.userinfo->privkey = argp->req.rec.userinfo->privkey;
    uac->ae.userinfo->srvprivkey = argp->req.rec.userinfo->srvprivkey;

    strbuf sb;
    sb << "Last modified " << timestr () << " by " << cp->unixcred->username;
    if (uid && !*uid) sb << "*superuser*";
    sb << "@";
    if (uid) sb << "LOCAL";
    else sb << client_name; // XXX do reverse DNS lookup on this ?
    uac->ae.userinfo->audit = sb;

    if (admin) {
      u_int32_t gid = argp->req.rec.userinfo->gid;
      if (udbp->gidmap && gid != udbp->gidmap->map (uac->ae.userinfo->gid)) {
	gid = udbp->gidmap->unmap (gid);
	if (gid == badid
	    || udbp->gidmap->map (gid) != argp->req.rec.userinfo->gid) {
	  *res.errmsg = "bad gid";
	  sbp->replyref (res);
	  return;
	}
      }
      uac->ae.userinfo->gid = gid;
      uac->ae.userinfo->privs = argp->req.rec.userinfo->privs;
    }

    if (!uac->update ()) {
      *res.errmsg = "database refused update";
      sbp->replyref (res);
      return;
    }

    res.set_ok (true);
  }
  else {
    *res.errmsg = "group updates not implemented";
    sbp->replyref (res);
    return;
  }

  sbp->replyref (res);
}

bool
authclnt::get_user_cursor (dbfile **dbpp, ptr<authcursor> *acp,
			   sfsauth_dbrec *dbrp, const sfsauth_dbkey &key,
			   bool writable)
{
  if (key.type != SFSAUTH_DBKEY_NAME && key.type != SFSAUTH_DBKEY_ID
      && key.type != SFSAUTH_DBKEY_PUBKEY) {
    if (dbrp) {
      dbrp->set_type (SFSAUTH_ERROR);
      *dbrp->errmsg = strbuf ("unsupported key type %d", key.type);
    }
    return false;
  }
  ptr<sfspub> pk;
  for (dbfile *dbp = dbfiles.base (); dbp < dbfiles.lim (); dbp++) {
    ptr<authcursor> ac = dbp->db->open (writable);
    if (!ac)
      continue;
    switch (key.type) {
    case SFSAUTH_DBKEY_NAME:
      {
	str name = dbp->strip_prefix (*key.name);
	if (!name || !ac->find_user_name (name)) {
	  struct passwd *pw;
	  if (dbp->allow_unix_pwd && (pw = getpwnam (name))) {
	    sfsauth_dbrec rec (SFSAUTH_USER);
	    rec.userinfo->name = *key.name;
	    rec.userinfo->id = pw->pw_uid;
	    rec.userinfo->vers = 0;
	    rec.userinfo->gid = pw->pw_gid;
	    ac->ae = rec;
	    break;
	  }
	  continue;
	}
	break;
      }
    case SFSAUTH_DBKEY_ID:
      {
	u_int32_t id = dbp->uidmap ? dbp->uidmap->unmap (*key.id) : *key.id;
	if (id == badid || !ac->find_user_uid (id)) {
	  struct passwd *pw;
	  if (dbp->allow_unix_pwd && (pw = getpwuid (*key.id))) {
	    sfsauth_dbrec rec (SFSAUTH_USER);
	    if (dbp->prefix)
	      rec.userinfo->name = dbp->prefix << "/" << pw->pw_name;
	    else
	      rec.userinfo->name = pw->pw_name;
	    rec.userinfo->id = pw->pw_uid;
	    rec.userinfo->vers = 0;
	    rec.userinfo->gid = pw->pw_gid;
	    ac->ae = rec;
	    break;
	  }
	  continue;
	}
	break;
      }
    case SFSAUTH_DBKEY_PUBKEY:
      if (!pk) {
	if (!(pk = sfscrpt.alloc (*key.key))) {
	  warn << "Cannot import user public key.\n";
	  return false;
	}
      }
      if (!ac->find_user_pubkey (*pk))
	continue;
      break;
    default:
      panic ("unreachable\n");
    }
    if (dbpp)
      *dbpp = dbp;
    if (acp)
      *acp = ac;
    if (dbrp) {
      *dbrp = ac->ae;
      dbrp->userinfo->pwauth = "";
      dbrp->userinfo->privkey.setsize (0);
      if (dbp->prefix)
	dbrp->userinfo->name = dbp->prefix << "/" << dbrp->userinfo->name;
      if (dbp->uidmap)
	dbrp->userinfo->id = dbp->uidmap->map (dbrp->userinfo->id);
      if (dbrp->userinfo->id == badid)
	continue;
      if (dbp->gidmap)
	dbrp->userinfo->gid = dbp->gidmap->map (dbrp->userinfo->gid);
    }
    return true;
  }
  if (dbrp) {
    dbrp->set_type (SFSAUTH_ERROR);
    *dbrp->errmsg = "not found";
  }
  return false;
}

void
authclnt::query_user (svccb *sbp)
{
  sfsauth2_query_arg *arg = sbp->template getarg<sfsauth2_query_arg> ();
  ptr<authcursor> ac;
  sfsauth2_query_res res;

  if (get_user_cursor (NULL, &ac, &res, arg->key)
      && sbp->getaui () < credtab.size ()) {
    const sfsauth_cred &c = credtab[sbp->getaui ()];
    // XXX - expanded for debug purposes.
    if (c.type == SFS_UNIXCRED &&
	c.unixcred->username == res.userinfo->name &&
	c.unixcred->uid == res.userinfo->id) 
      res.userinfo->privkey = ac->ae.userinfo->privkey;
  }
  sbp->replyref (res);
}

void
authclnt::query_srpparms (svccb *sbp)
{
  str srpfile, parms;
  sfsauth2_query_res res (SFSAUTH_SRPPARMS);
  if (!(srpfile = sfsconst_etcfile ("sfs_srp_parms")) |
      !(parms = file2str (srpfile))) {
    res.set_type (SFSAUTH_ERROR);
    *res.errmsg = "No SRP information available";
  } else {
    res.srpparms->parms = parms;
  }
  sbp->replyref (res);
}

void
authclnt::query_certinfo (svccb *sbp)
{
  sfsauth2_query_res res (SFSAUTH_CERTINFO);
  if (sfsauthrealm.len () > 0) {
    res.certinfo->name = sfsauthrealm;
    res.certinfo->info.set_status (SFSAUTH_CERT_REALM);
    res.certinfo->info.certpaths->set (sfsauthcertpaths.base (), 
				       sfsauthcertpaths.size (), 
				       freemode::NOFREE);
  }
  else {
    res.certinfo->name = myservinfo.cr7->host.hostname;
    res.certinfo->info.set_status (SFSAUTH_CERT_SELF);
  }
  sbp->replyref (res);
  return ;

}

void
authclnt::query_group (svccb *sbp)
{
  sfsauth2_query_arg *arg = sbp->template getarg<sfsauth2_query_arg> ();
  sfsauth2_query_res res;
  if (arg->key.type != SFSAUTH_DBKEY_NAME
      && arg->key.type != SFSAUTH_DBKEY_ID) {
    res.set_type (SFSAUTH_ERROR);
    *res.errmsg = strbuf ("unsupported key type %d", arg->key.type);
    sbp->replyref (res);
    return;
  }
  for (dbfile *dbp = dbfiles.base (); dbp < dbfiles.lim (); dbp++) {
    ptr<authcursor> ac = dbp->db->open ();
    if (!ac)
      continue;
    switch (arg->key.type) {
    case SFSAUTH_DBKEY_NAME:
      {
	str name = dbp->strip_prefix (*arg->key.name);
	if (!name || !ac->find_group_name (name))
	  continue;
	break;
      }
    case SFSAUTH_DBKEY_ID:
      {
	u_int32_t id = dbp->uidmap ? dbp->uidmap->unmap (*arg->key.id)
	  : *arg->key.id;
	if (id == badid || !ac->find_group_gid (id))
	  continue;
	break;
      }
    default:
      panic ("unreachable");
    }
    res = ac->ae;
    assert (res.type == SFSAUTH_GROUP);
    sbp->replyref (res);
    return;
  }
  res.set_type (SFSAUTH_ERROR);
  *res.errmsg = "not found";
  sbp->replyref (res);
}


void
authclnt::sfsauth_query (svccb *sbp)
{
  sfsauth2_query_arg *arg = sbp->template getarg<sfsauth2_query_arg> ();
  switch (arg->type) {
  case SFSAUTH_USER:
    query_user (sbp);
    break;
  case SFSAUTH_GROUP:
    query_group (sbp);
    break;
  case SFSAUTH_CERTINFO:
    query_certinfo (sbp);
    break;
  case SFSAUTH_SRPPARMS:
    query_srpparms (sbp);
    break;
  default:
    sfsauth2_query_res res;
    res.set_type (SFSAUTH_ERROR);
    *res.errmsg = strbuf ("unsupported query type %d", arg->type);
    sbp->replyref (res);
    break;
  }
}

void
authclnt::sfs_login (svccb *sbp)
{
  if (!authid_valid) {
    sbp->replyref (sfs_loginres (SFSLOGIN_ALLBAD));
    return;
  }
  sfsauth2_loginarg la;
  la.arg = *sbp->template getarg<sfs_loginarg> ();
  la.authid = authid;
  la.source = client_name;

  sfsauth2_loginres lr;
  sfsauth_login (&lr, &la, true);

  sfs_loginres res (lr.status);
  switch (lr.status) {
  case SFSLOGIN_OK:
    if (!seqstate.check (la.arg.seqno)
	|| lr.resok->creds.size () < 1)
      res.set_status (SFSLOGIN_BAD);
    else {
      *res.authno = authalloc (lr.resok->creds.base(), 
			       lr.resok->creds.size ());
      if (!*res.authno) {
	warn << "ran out of authnos (or bad cred type)\n";
	res.set_status (SFSLOGIN_BAD);
      }
    }
    break;
  case SFSLOGIN_MORE:
    *res.resmore = *lr.resmore;
    break;
  default:
    break;
  }

  sbp->replyref (res);
}

void
authclnt::sfsauth_login (sfsauth2_loginres *resp, const sfsauth2_loginarg *lap,
			 bool srp_ok)
{
  sfs_autharg2 aa;
  sfsauth2_loginres &res = *resp;
  if (!bytes2xdr (aa, lap->arg.certificate)) {
    res.set_status (SFSLOGIN_BAD);
    *res.errmsg = "cannot unmarshal certificate";
    return;
  }

  switch (aa.type) {
  case SFS_AUTHREQ:
  case SFS_AUTHREQ2:
    authreq (&res, lap, &aa);
    break;
  case SFS_UNIXPWAUTH:
    unixpwauth (&res, lap, &aa, srp_ok);
    break;
  case SFS_SRPAUTH:
    if (srp_ok)
      srpauth (&res, lap, &aa);
    else {
      res.set_status (SFSLOGIN_BAD);
      *res.errmsg = "SRP authentication of client to third party not allowed";
    }
    break;
  default:
    res.set_status (SFSLOGIN_BAD);
    *res.errmsg = strbuf ("unknown login type %d", aa.type);
    break;
  }
}

void
authclnt::authreq (sfsauth2_loginres *resp,
		   const sfsauth2_loginarg *lap, const sfs_autharg2 *aap)
{
  ptr<sfspub> vrfy;
  str logname;
  if (aap->type == SFS_AUTHREQ) {
    const sfs_pubkey &kp = aap->authreq1->usrkey;
    if (!(vrfy = sfscrpt.alloc (kp, SFS_VERIFY))) {
      *resp->errmsg = "Cannot load public Rabin key";
      resp->set_status (SFSLOGIN_BAD);
      return ;
    }
    sfs_signed_authreq authreq;
    str msg;
    if (!vrfy->verify_r (aap->authreq1->signed_req, sizeof (authreq), msg)
	|| !str2xdr (authreq, msg)
	|| authreq.type != SFS_SIGNED_AUTHREQ
	|| authreq.seqno != lap->arg.seqno
	|| authreq.authid != lap->authid) {
      resp->set_status (SFSLOGIN_BAD);
      *resp->errmsg = "bad signature";
      return;
    }
    if (authreq.usrinfo[0]) {
      if (memchr (authreq.usrinfo.base (), 0, authreq.usrinfo.size ()))
	logname = authreq.usrinfo.base ();
      else
	logname.setbuf (authreq.usrinfo.base (), authreq.usrinfo.size ());
    }
  }
  else {
    if (aap->sigauth->req.user.len ())
      logname = aap->sigauth->req.user;
    str msg = xdr2str (aap->sigauth->req);
    strbuf err;
    resp->set_status (SFSLOGIN_BAD);
    if (!(vrfy = sfscrpt.alloc (aap->sigauth->key, SFS_VERIFY))) {
      *resp->errmsg = "cannot load public key";
      return; 
    } else if (!vrfy->verify (aap->sigauth->sig, msg)) {
      *resp->errmsg = "Bad login: Could not verify signature\n";
      return;
    }
    if (!checkreq (resp, lap, &aap->sigauth->req))
      return;
  }
  for (dbfile *dbp = dbfiles.base (); dbp < dbfiles.lim (); dbp++) {
    ptr<authcursor> ac = dbp->db->open ();
    if (ac && ac->find_user_pubkey (*vrfy) && ac->ae.type == SFSAUTH_USER
	&& (!logname || (!dbp->prefix && logname == ac->ae.userinfo->name)
	    || logname == dbp->prefix << "/" << ac->ae.userinfo->name)
	&& setuser (resp, ac->ae, dbp)) {
      str source = "";
      if (lap->source.len ())
	source = strbuf () << " from " << lap->source;
      str name = ac->ae.userinfo->name;
      if (dbp->prefix)
	name = dbp->prefix << "/" << name;
      warn << "signed login by " << name << source << "\n";
      return;
    }
  }
  resp->set_status (SFSLOGIN_BAD);
  *resp->errmsg = "bad login";
}

#define ROOT_SU(pw, ok)   (!*(pw) && (ok) && uid && *uid == 0)

void
authclnt::unixpwauth (sfsauth2_loginres *resp, const sfsauth2_loginarg *lap,
		      const sfs_autharg2 *aap, bool root_su_ok)
{
  dbfile *dbp = dbfiles.base ();
  str name, source;
  ptr<authcursor> ac;
  bool flag = false;

  struct passwd *pe = getpwnam (aap->pwauth->req.user.cstr ());
  const char *pw;
  if (!pe) 
    goto badlogin;
  pw = aap->pwauth->password.cstr ();

  while (dbp < dbfiles.lim ()) {
    if (dbp->allow_unix_pwd) {
      flag = true;
      break;
    }
    dbp++;
  }
  if (!flag)
    goto badlogin;

#if HAVE_GETSPNAM
  if (struct spwd *spe = getspnam (aap->pwauth->req.user.cstr ())) {
    if (!ROOT_SU (pw, root_su_ok) &&
	strcmp (spe->sp_pwdp, crypt (pw, spe->sp_pwdp)))
      goto badlogin;
  }
  else
#endif /* HAVE_GETSPNAM */
    if (!ROOT_SU (pw, root_su_ok) &&
	strcmp (pe->pw_passwd, crypt (pw, pe->pw_passwd)))
      goto badlogin;

  ac = dbp->db->open ();
  if (!ac)
    goto badlogin;
  if (!ac->find_user_name (pe->pw_name)) {
    ac->ae.set_type (SFSAUTH_USER);
    ac->ae.userinfo->name = pe->pw_name;
    ac->ae.userinfo->id = pe->pw_uid;
    ac->ae.userinfo->vers = 0;
    ac->ae.userinfo->gid = pe->pw_gid;
  }

  if (!setuser (resp, ac->ae, dbp))
    return;
  source = "";
  if (lap->source.len ())
    source = strbuf () << " from " << lap->source;
  name = ac->ae.userinfo->name;
  if (dbp->prefix)
    name = dbp->prefix << "/" << name;
  warn << "password login by " << name << source << "\n";
  return;

 badlogin:
  resp->set_status (SFSLOGIN_BAD);
  *resp->errmsg = "bad login";
}

void
authclnt::srpauth (sfsauth2_loginres *resp, const sfsauth2_loginarg *lap,
		   const sfs_autharg2 *aap)
{
  if (aap->srpauth->req.authid != lap->authid
      || (!sfsauth_login_srp && (!authid_valid || lap->authid != authid))) {
    srp_ac = NULL;
    resp->set_status (SFSLOGIN_BAD);
    *resp->errmsg = "invalid or missing authid";
    return;
  }
  if (aap->srpauth->msg.size () == 0) {
    sfsauth_dbkey kname (SFSAUTH_DBKEY_NAME);
    *kname.name = aap->srpauth->req.user;
    if (!get_user_cursor (&srp_dbp, &srp_ac, NULL, kname)) {
      srp_ac = NULL;
      resp->set_status (SFSLOGIN_BAD);
      *resp->errmsg = "bad login";
      return;
    }
    srp_seqno = lap->arg.seqno;
    srp_authid = lap->authid;
    resp->set_status (SFSLOGIN_MORE);
    switch (srp.init (resp->resmore.addr (), &aap->srpauth->msg,
		      srp_authid, srp_ac->ae.userinfo->name,
		      srp_ac->ae.userinfo->pwauth)) {
    case SRP_NEXT:
      return;
    default:
      srp_ac = NULL;
      resp->set_status (SFSLOGIN_BAD);
      *resp->errmsg = "bad login";
      return;
    }
  }
  if (!srp_ac || srp_authid != lap->authid || srp_seqno != lap->arg.seqno) {
    resp->set_status (SFSLOGIN_BAD);
    *resp->errmsg = "unknown SRP session";
    return;
  }
  srpmsg msg;
  switch (srp.next (&msg, &aap->srpauth->msg)) {
  case SRP_NEXT:
    resp->set_status (SFSLOGIN_MORE);
    *resp->resmore = msg;
    return;
  case SRP_LAST:
    if (setuser (resp, srp_ac->ae, srp_dbp)) {
      resp->resok->resmore = msg;
    }
   break;
  default:
    resp->set_status (SFSLOGIN_BAD);
    *resp->errmsg = "bad login";
    break;
  }
  srp_ac = NULL;
}
