/*
 *
 * Copyright (C) 2004 Alexander Yip (yipal@mit.edu),
 *                    Massachusetts Institute of Technology
 * 
 *
 *  Permission is hereby granted, free of charge, to any person obtaining
 *  a copy of this software and associated documentation files (the
 *  "Software"), to deal in the Software without restriction, including
 *  without limitation the rights to use, copy, modify, merge, publish,
 *  distribute, sublicense, and/or sell copies of the Software, and to
 *  permit persons to whom the Software is furnished to do so, subject to
 *  the following conditions:
 *
 *  The above copyright notice and this permission notice shall be
 *  included in all copies or substantial portions of the Software.
 *
 *  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 *  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 *  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 *  NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
 *  LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
 *  OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
 *  WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 *
 */
#include "merkle_misc.h"
#include "dhash_blockstore.h"
#include "misc_utils.h"
#include "db_wrap.h"
#include "verify.h"
#include <db.h>

dhash_blockstore::dhash_blockstore() {
  block_db.init();
  keyhash_db.init();
}
dhash_blockstore::~dhash_blockstore() {
  block_db.close();
  keyhash_db.close();
  if(dbe) dbe->close(dbe, 0);
}


int dhash_blockstore::open_db(DB **db, DB_ENV *dbe, char *filename, int preflags, int flags) {
  int r;
  int mode = 0664;

  r = db_create(db, dbe, 0);
  if (r) return r;;

  if (preflags != 0) {
    r = (*db)->set_flags((*db), preflags);
    if (r) return r;
  }

  (*db)->set_pagesize((*db), 16 * 1024);

#if ((DB_VERSION_MAJOR < 4) || ((DB_VERSION_MAJOR == 4) && (DB_VERSION_MINOR < 1)))
  r = open_database (*db, NULL, filename, NULL, DB_BTREE, flags, mode);
#else
  r = open_database (*db, NULL, filename, NULL, DB_BTREE, flags|DB_AUTO_COMMIT, mode);
#endif

  return r;
}

int dhash_blockstore::dbrec2dbt(ptr<dbrec> src, DBT *dst, bigint *projkey) {
  // <dst> must point to allocated memory
  // dst->data must be cleaned up by caller
  bzero(dst, sizeof(*dst));
  if (projkey) {
    strbuf sb = strbuf() << bigint2rawstr(*projkey) << str(src->value, src->len);
    dst->data = malloc(src->len + 20);
    memcpy(dst->data, str(sb).cstr(), src->len + 20);
    dst->size = src->len + 20;
  } else {
    dst->size = src->len;
    dst->data = malloc(src->len);
    memcpy(dst->data, src->value, src->len);
  }
  return 0;
}

ptr<dbrec> dhash_blockstore::dbt2dbrec(DBT *src, bigint *projkey) {
  if (projkey) {
    assert(src->size-20 >= 0);
    str s = substr(str((char*)src->data, src->size), 0, 20);
    mpz_set_rawmag_be(projkey, s.cstr(), s.len());
    return New refcounted<dbrec>((char*)src->data+20, src->size-20);
  } else {
    return New refcounted<dbrec>(src->data, src->size);
  }
}

int dhash_blockstore::bigint2dbt(bigint src, DBT *dst) {
  // dst->data must be cleaned up by caller
  bzero(dst, sizeof(*dst));
  dst->data = malloc(20);
  mpz_get_rawmag_be((char*)dst->data, 20, &src);
  dst->size = 20;
  return 0;
}

ptr<dbrec> dhash_blockstore::bigint2dbrec(bigint src) {
  char buf[20];
  mpz_get_rawmag_be(buf, sizeof(buf), &src);
  return New refcounted<dbrec>(buf, sizeof(buf));
}

int dhash_blockstore::insert_block(DB *db, DB_TXN *txn, bigint *projkey, 
				   ptr<dbrec> key, ptr<dbrec> data) {
  int r;
  DBT skey, sdata;
  dbrec2dbt(key, &skey);
  dbrec2dbt(data, &sdata, projkey);
  
  r = db->put(db, txn, &skey, &sdata, 0);
  free(skey.data);
  free(sdata.data);
  return r;
}

int dhash_blockstore::insert_trackedblock(db_agg *agg, bigint projkey, 
					  ptr<dbrec> key, ptr<dbrec> data){
  DB_TXN *txn = NULL;
  int r;
  ptr<dbrec> projkeyrec = bigint2dbrec(projkey);
  ptr<dbrec> metadbrec;
  metadata_entry e;

  r = begin_transaction (dbe, NULL, &txn, 0);
  if (r) return r;
  r = insert_block(agg->data, txn, &projkey, key, data);
  if (r) goto end;
  r = insert_block(agg->proj2blocks, txn, NULL, projkeyrec, key);
  if (r) goto end;
  e.setprojkey(projkey);
  metadbrec = New refcounted<dbrec>(&e, sizeof(e));
  //warn << "size of metadata: " << sizeof(e) << " " << metadbrec->len << "\n";
  r = insert_block(agg->metadata, txn, NULL, key, metadbrec);
  if (r) goto end;

 end:
  if (r) {
    warn << "ABORTING INSERT!! " << armor64A(dbrec2str(key)) << "\n";
    return abort_transaction (txn);
  }
  return commit_transaction (txn, 0);
}

int 
dhash_blockstore::getblockids_from(db_agg *agg, vec < ptr<dbrec> > *rec_list, 
				   bigint *projkey) {
  int r;
  DBC *cursor;
  assert(agg);
  rec_list->clear();

  if (projkey) {
    // return keys in proj2blocks pointed to by <projkey>
    DBT matchkey, data;
    bzero(&data, sizeof(data));
    bigint2dbt(*projkey, &matchkey);

    // Need to remember matchpointer, because c_get may change matchkey.data
    void *matchpointer = matchkey.data; 

    r = agg->proj2blocks->cursor(agg->proj2blocks, NULL, &cursor, 0);
    assert(!r);

    for (r = cursor->c_get(cursor, &matchkey, &data, DB_SET);
	 r == 0;
	 r = cursor->c_get(cursor, &matchkey, &data, DB_NEXT_DUP))
      rec_list->push_back( dbt2dbrec(&data) );
    free(matchpointer);

  } else {
    // return keys of all records in db
    DBT key, data;
    bzero(&key, sizeof(key));
    bzero(&data, sizeof(data));

    r = agg->metadata->cursor(agg->metadata, NULL, &cursor, 0);
    if (r) goto end;
    
    while ((r = cursor->c_get(cursor, &key, &data, DB_NEXT)) == 0)
      rec_list->push_back( dbt2dbrec(&key) );
  }
  
 end:
  r = cursor->c_close(cursor);

  if (r && r != DB_NOTFOUND) {
    warn << "getblockids_from: " << db_strerror(r) << "\n";
    return r;
  }
  return 0;
}

static bool between_inclusive (const chordID &a, const chordID &b, const chordID &n)
{
  bool r;
  if (a == b) {
    r = n == a;  
  } else if (a < b) {
    r = (n >= a) && (n <= b);
  } else {
    r = (n >= a) || (n <= b);
  }
  return r;
}


int 
dhash_blockstore::getchblockids(vec<ptr<dbrec> > *rec_list, 
				ptr<dbrec> start, ptr<dbrec> end, uint32_t max) {
  int r;
  unsigned int count = 0;
  DBC *cursor;
  rec_list->clear();
  chordID current, low, high;
  low  = dbrec2id(start);
  high = dbrec2id(end);

  DBT key, data;
  bzero(&data, sizeof(data));
  dbrec2dbt(start, &key);
  void *freeme = key.data; 

  r = block_db.data->cursor(block_db.data, NULL, &cursor, 0);
  if (r) goto end;

  // before ring origin
  r = cursor->c_get(cursor, &key, &data, DB_SET_RANGE);
  while(!r) {
    current = dbrec2id( dbt2dbrec(&key));
    if (max && count == max) break;
    if (end && !between_inclusive(low, high, current)) break;

    rec_list->push_back( dbt2dbrec(&key) );
    r = cursor->c_get(cursor, &key, &data, DB_NEXT);
    count++;
  }
  free(freeme);

  // after ring origin
  bigint2dbt(bigint(0), &key);
  freeme = key.data;
  r = cursor->c_get(cursor, &key, &data, DB_SET_RANGE);
  while(!r) {
    current = dbrec2id( dbt2dbrec(&key));
    if (max && count == max) break;
    if (end && !between_inclusive(low, high, current)) break;
    if (current >= low) break;

    rec_list->push_back( dbt2dbrec(&key) );
    r = cursor->c_get(cursor, &key, &data, DB_NEXT);
    count++;
  }
  free(freeme);

  r = cursor->c_close(cursor);

 end:
  if (r && r != DB_NOTFOUND) {
    warn << "getblockids: " << db_strerror(r) << "\n";
    return r;
  }
  
  return 0;
}

ptr<dbrec> 
dhash_blockstore::lookup_block(DB *db, ptr<dbrec> key, bigint *projkey){
  bigint keycontent;
  DBT skey, content;
  bzero(&skey, sizeof(skey));
  skey.size = key->len;
  skey.data = key->value;
  bzero(&content, sizeof(content));
  int r=0;

  if ((r = db->get(db, NULL, &skey, &content, 0)) != 0) 
    return NULL;

  if (projkey) 
    return dbt2dbrec(&content, projkey);
  else
    return dbt2dbrec(&content, &keycontent);
}

int
dhash_blockstore::remove_trackedblock(db_agg *agg, ptr<dbrec> key) {
  DB_TXN *txn = NULL;
  DBT datakey, content, backptrkey;
  DBC *cursor;
  int r;
  bigint projkey;
  dbrec2dbt(key, &datakey);
  void *freeme = datakey.data;
  bzero(&content, sizeof(content));

  r = begin_transaction (dbe, NULL, &txn, 0);
  if (r) goto end;

  // extract projkey and convert to DBT
  r = agg->data->get(agg->data, txn, &datakey, &content, 0); 
  if (r) goto end;

  dbt2dbrec(&content, &projkey);
  bigint2dbt(projkey, &backptrkey); 
  
  // remove proj2blocks
  cursor = NULL;
  r = agg->proj2blocks->cursor(agg->proj2blocks, txn, &cursor, 0);
  if (r) goto end;
  r = cursor->c_get(cursor, &backptrkey, &datakey, DB_GET_BOTH);
  if (r) goto end;
  r = cursor->c_del(cursor, 0);
  if (r) goto end;
  r = cursor->c_close(cursor);
  if (r) goto end;

  // remove metadata
  r = agg->metadata->del (agg->metadata, txn, &datakey, 0);
  if (r) goto end;

  // remove data block
  r = agg->data->del (agg->data, txn, &datakey, 0);
  if (r) goto end;

 end:
  free(freeme);
  if(txn) {
    if (r)
      abort_transaction (txn);
    else
      r = commit_transaction (txn, 0);
  }
  if (r)
    warn << "remove_trackedblock: " << db_strerror(r) << "\n";

  return r;
}

// -------------------------------------------------------------------
int dhash_blockstore::opendb(str dirname, bool createenv, time_t synctime) {
  char cpath[MAXPATHLEN];
  int fd = -1;
  int r = -1;
  long cacheSize = 1000 * 1024;
  int envflags = 0;
  synctime_ = synctime;

  fd = open(".", O_RDONLY);
  if (fd < 0) fatal << "can't open current working dir\n";
  mkdir(dirname, 0755);
  r = chdir(dirname);
  if (r == -1) fatal << "couldn't chdir to " << dirname << "\n";

  // Create/Join DB environment
  r = db_env_create(&dbe, 0);
  if (r) goto end;
  if (cacheSize > 0) {
    r = dbe->set_cachesize(dbe, 0, cacheSize, 0);
    if (r) goto end;
  }

  getcwd(cpath, MAXPATHLEN);
  r = dbe->set_data_dir(dbe, cpath);
  if (r) goto end;
  r = dbe->set_lg_dir(dbe, cpath);
  if (r) goto end;
  r = dbe->set_tmp_dir(dbe, cpath);
  if (r) goto end;
  envflags = (createenv) ? (DB_CREATE | DB_INIT_MPOOL | DB_INIT_LOCK |
			    DB_INIT_LOG | DB_INIT_TXN | DB_RECOVER)  : (DB_JOINENV);
  r = dbe->open(dbe, NULL, envflags, 0);
  if (r) goto end;
  dbe->set_errfile(dbe, stderr);

#if ((DB_VERSION_MAJOR > 4) || ((DB_VERSION_MAJOR == 4) && (DB_VERSION_MINOR >= 2)))
  dbe->set_flags(dbe, DB_LOG_AUTOREMOVE, 1);
#endif

  // Create/Open databases
  open_db(&block_db.data, dbe, "blocks", 0, DB_CREATE);
  open_db(&block_db.proj2blocks, dbe, "block_proj2blocks", DB_DUP | DB_DUPSORT, DB_CREATE);
  open_db(&block_db.metadata, dbe, "block_metadata", 0, DB_CREATE);

  open_db(&keyhash_db.data, dbe, "keyhash", 0, DB_CREATE);
  open_db(&keyhash_db.proj2blocks, dbe, "keyhash_proj2blocks", DB_DUP | DB_DUPSORT, DB_CREATE);
  open_db(&keyhash_db.metadata, dbe, "keyhash_metadata", 0, DB_CREATE);

 end:
  fchdir(fd);
  close(fd);

  if (synctime)
    delaycb (synctime, wrap (this, &dhash_blockstore::sync_cb));

  if (r)
    warn << "dhash_blockstore::opendb: " << db_strerror(r) << "\n";
  return r;
}

void dhash_blockstore::sync_cb() {
  block_db.sync();
  keyhash_db.sync();
  checkpoint(dbe, 1024, 0, 0);
  delaycb (synctime_, wrap (this, &dhash_blockstore::sync_cb));
}

int 
dhash_blockstore::insert_pkblock(bigint projkey, ptr<dbrec> key, ptr<dbrec> data){
  //warn << "Inserting PK block: " << armor64A(projkey) << " " << armor64A(dbrec2str(key)) << " " << dbrec2id(key) << "\n";
  return insert_trackedblock(&keyhash_db, projkey, key, data);
}
int 
dhash_blockstore::insert_chblock(bigint projkey, ptr<dbrec> key, ptr<dbrec> data){
  //warn << "Inserting CH block: " << armor64A(projkey) << " " << armor64A(dbrec2str(key)) << " " << dbrec2id(key) << "\n";
  return insert_trackedblock(&block_db, projkey, key, data);
}
int 
dhash_blockstore::remove_pkblock(ptr<dbrec> key){
  //warn << "Removing  PK block: "  << " " << armor64A(dbrec2str(key)) << " " << dbrec2id(key) << "\n";
  return remove_trackedblock(&keyhash_db, key);
}
int 
dhash_blockstore::remove_chblock(ptr<dbrec> key){
  //warn << "Removing  CH block: "  << " " << armor64A(dbrec2str(key)) << " " << dbrec2id(key) << "\n";
  return remove_trackedblock(&block_db, key);
}
ptr<dbrec> 
dhash_blockstore::lookup_pkblock(ptr<dbrec> key, bigint *projkey){
  return lookup_block(keyhash_db.data, key, projkey);
}
ptr<dbrec> 
dhash_blockstore::lookup_chblock(ptr<dbrec> key, bigint *projkey){
  return lookup_block(block_db.data, key, projkey);
}
int
dhash_blockstore::ch_get_projkey(ptr<dbrec> key, bigint *projkey) {
  DBT tkey, pkey;
  bzero(&tkey, sizeof(tkey));
  tkey.size = key->len;
  tkey.data = key->value;
  bzero(&pkey, sizeof(pkey));
  metadata_entry e;
  int r = 0;

  if ((r = block_db.metadata->get(block_db.metadata, NULL, &tkey, &pkey, 0)) != 0)
    return r;
  
  assert (pkey.size == sizeof(e));
  memcpy (&e, pkey.data, pkey.size);
  *projkey = e.getprojkey();
  return 0;
}
int 
dhash_blockstore::getpkblockids(vec<ptr<dbrec> > *rec_list, bigint *projkey) {
  return getblockids_from(&keyhash_db, rec_list, projkey);
}
int
dhash_blockstore::getchblockids(vec<ptr<dbrec> > *rec_list, bigint *projkey) {
  return getblockids_from(&block_db, rec_list, projkey);
}

bsEnumeration::bsEnumeration(DB* db) {
  int r;
  r = db->cursor(db, NULL, &cursor, 0);
  assert (!r);
}
bsEnumeration::~bsEnumeration() {
  if (cursor)
    close();
}
void
bsEnumeration::close() {
  cursor->c_close(cursor);
}

ptr<dbrec>
bsEnumeration::nextElement() {
  int r;
  DBT key, data;
  bzero(&key, sizeof(key));
  bzero(&data, sizeof(data));

  ptr<dbrec> keyrec;
  r = cursor->c_get(cursor, &key, &data, DB_NEXT);
  if (r) goto end;
  keyrec = New refcounted<dbrec>(key.data, key.size);

 end:
  if (r) {
    if (r != DB_NOTFOUND) warn << "error in nextElement: " << db_strerror(r) << "\n";
    return NULL;
  }
  return keyrec;
}

ptr<dbrec>
bsEnumeration::nextElement(ptr<dbrec> startkey) {
  int r;
  DBT key, data;
  bzero(&key, sizeof(key));
  bzero(&data, sizeof(data));

  key.size = startkey->len;
  key.data = startkey->value;

  ptr<dbrec> keyrec;
  r = cursor->c_get(cursor, &key, &data, DB_SET_RANGE);
  if (r) goto end;
  keyrec = New refcounted<dbrec>(key.data, key.size);

 end:
  if (r) {
    if (r != DB_NOTFOUND) warn << "error in nextElement: " << db_strerror(r) << "\n";
    return NULL;
  }
  return keyrec;
}

void dhash_blockstore::dump_pk(str message, bigint *filterkey) {
  warn << message;
  bigint projkey;
  unsigned int i;
  vec<ptr<dbrec> > rec_list;
  getblockids_from(&keyhash_db, &rec_list, filterkey);
  
  for(i=0; i<rec_list.size(); i++) {
    ptr<dbrec> data = lookup_block(keyhash_db.data, rec_list[i], &projkey);
    long version = keyhash_version (data);
    warn << "    " << armor64A(projkey)
      	 << "  " << armor64A(dbrec2id(rec_list[i]))
	 << " 0x" << hexdump(rec_list[i]->value, rec_list[i]->len) 
	 << " ver " << version 
	 << "\n";
  }
}
void dhash_blockstore::dump_ch(str message, bigint *filterkey) {
  warn << message;
  bigint projkey;
  unsigned int i;
  vec<ptr<dbrec> > rec_list;
  getblockids_from(&block_db, &rec_list, filterkey);
  
  for(i=0; i<rec_list.size(); i++) {
    lookup_block(block_db.data, rec_list[i], &projkey);
    warn << "    " << armor64A(projkey)
	 << "  " << armor64A(dbrec2id(rec_list[i]))
	 << " 0x" << hexdump(rec_list[i]->value, rec_list[i]->len) << "\n";
  }
}

