/*
 *  Copyright (C) 2002-2004  Massachusetts Institute of Technology
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License as
 *  published by the Free Software Foundation; either version 2, or (at
 *  your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 *  USA
 */

#include <dhash_common.h>
#include <dhashclient.h>
#include <verify.h>
#include "id_utils.h"

#include "stdinc.h"
#include "ns.h"
#include "pastdbfe.h"
#include "lrucache.h"
#include "record.h"
#include "penviron.h"

static int rpc_window = atoi(pgetenv("CFSRW_RPCWIN"));
static int rpc_timer = atoi(pgetenv("CFSRW_RPCTIMER"));
static int bytes_downloaded = 0;
extern int ns_debug;

struct pk_insert_state {
  ptr<sfspriv> sk;
  str data;
  long version;
  cbinsert cb;
  pk_insert_state (ptr<sfspriv> k, str d, cbinsert c)
    : sk (k), data (d), cb (c) {}
};

class ns_impl : public virtual ns {
private:
  bigint _project;
  bool _connected;

  ptr<dhashclient> _dhc;
  ptr<pastdbfe> _bdb;
  str _bdbf;
  vec<bigint> _stalepks;

  // These are version numbers at the dhash level, not to be confused with
  // version numbers at the pastwatch level.
  lrucache<bigint, long, hashID> _pkv; 

public:
  ns_impl (bigint n); 
  ~ns_impl ();
  void exit_cleanup ();
 
  void sync () { _bdb->sync (); }
  void connect (cbv);
  void connect (str, cbv);
  bool connected () const { return _connected; }

  bigint insert_local (const char *, size_t);
  ptr<ns_block> retrieve_local (bigint n);
  bool check_local (bigint n);

  bigint insert_block (const char *, size_t, cbinsert cb, bool nocache = false);
  void retrieve_block (bigint n, cbblock cb, bool nocache = false);

  void insert_pk (ptr<sfspriv> sk, str data, cbinsert cb);
  void insert_raw_pk_from_cache (bigint n, cbinsert cb);

  void retrieve_pk (bigint n, cbkeyhash cb);
  void retrieve_pk_from_net (bigint n, cbkeyhash cb);
  void checkpks (vec<bigint> n, vec<long> v, cbcheckpks cb);

  // helpers
  void retrieve_lptr (bigint n, cblptr cb);
  void retrieve_commit (bigint n, cbcrec cb);

private:
  int _fctr;
  vec<bool> _isinsert;
  vec<bigint> _ibid;
  vec<dhash_ctype> _itype;
  vec<ptr<ns_block> > _iblock;
  vec<ptr<option_block> > _iopt;
  vec<cbinsertgw_t> _icbs;
  vec<bigint> _rbid;
  vec<dhash_ctype> _rtype;
  vec<cb_cret> _rcbs;
  vec<ptr<option_block> > _ropt;
  unsigned _timedoutrpcs;
  
  void rpc_timed_out ();
  void run_next ();
  void next ();
  void got_connection (cbv cb, ptr<axprt_stream> x);
  void insert_cb (ptr<bool> exp, cbinsertgw_t cb,
                  dhash_stat status, ptr<insert_info> i);
  void insert_timedout (ptr<bool> exp, bigint n, cbinsertgw_t cb);
  void retrieve_cb (ptr<bool> exp, cb_cret cb,
                    dhash_stat status, ptr<dhash_block> b, vec<chordID> rt);
  void retrieve_timedout (ptr<bool> exp, cb_cret cb);
  void retrieve_block_cb
    (cbblock cb, bigint n, bool skipc,
     dhash_stat status, ptr<dhash_block> b, vec<chordID> rt);
  void insert_block_cb
    (bool nocache, cbinsert cb, dhash_stat status, ptr<insert_info> i);
  void insert_pk_cb (ptr<bool> exp, ptr<pk_insert_state> st,
                     dhash_stat status, ptr<insert_info> i);
  void insert_pk_timedout (ptr<bool> exp, bigint n, ptr<pk_insert_state> st);
  void insert_pk (ptr<pk_insert_state> st);
  void fetch_pk_for_insert
    (bigint n, ptr<pk_insert_state> st, ptr<keyhash_payload>);
  void retrieve_pk_cb
    (bigint n, ptr<dhash_block> cached, cbkeyhash cb,
     dhash_stat status, ptr<dhash_block> b, vec<chordID> rt);
  void retrieve_pk_timedout (ptr<bool> exp, cb_cret cb);
  void retrieve_pk_ret
    (ptr<bool> exp, cb_cret cb,
     dhash_stat status, ptr<dhash_block> b, vec<chordID> rt);
  void retrieve_pk_nc_cb
    (bigint n, cbkeyhash cb,
     dhash_stat status, ptr<dhash_block> b, vec<chordID> rt);
  void checkpks_cb (ptr<bool> exp, cbcheckpks cb,
                    dhash_stat status, vec<bigint> n, vec<long> v);
  void check_pks_timedout (ptr<bool> exp, cbcheckpks cb);
};

ptr<ns>
ns::alloc (bigint v)
{
  return New refcounted<ns_impl> (v);
}

void
ns_impl::rpc_timed_out ()
{
  _timedoutrpcs++;
  if (_timedoutrpcs > 4)
    _connected = false;
}

extern void
make_connection (const char *args, callback<void,ptr<axprt_stream> >::ptr cb);

void
ns_impl::got_connection (cbv cb, ptr<axprt_stream> x)
{
  if (x) {
    _dhc = New refcounted<dhashclient> (x);
    _connected = true;
  }
  else {
    warn << "cannot connect to any DHT host\n";
    _dhc = 0;
    _connected = false;
  }
  cb ();
}

void
ns_impl::connect (cbv cb)
{
  char *sn = pgetenv ("PASTWATCH_TCPSOCKET");
  if (!sn) {
    warn << "cannot connect to any DHT host\n";
    _dhc = 0;
    _connected = false;
    cb ();
  }
  else
    connect (sn, cb);
}

void
ns_impl::connect (str sn, cbv cb)
{
  assert (_dhc == 0);
  assert (!_connected);
  make_connection (sn, wrap (this, &ns_impl::got_connection, cb));
}

ns_impl::ns_impl (bigint n)
  : _project (n), _pkv (256)
{
  _connected = false;
  _dhc = 0;
  _bdb = 0;
  _bdbf = armor64A (_project);
  _timedoutrpcs = 0;
  _fctr = 0;

  str p = pget_replicadir() << "/" << _bdbf;
  str b = p << "/blocks.db";
  mkdir (p, 0755);
  
  dbOptions opts;
  opts.addOption("opt_flag", DB_CREATE);
  opts.addOption ("opt_cachesize", 1000);
  opts.addOption ("opt_nodesize", 4096);
  // opts.addOption ("opt_join", 1);

  _bdb = New refcounted<pastdbfe> ();
  if (_bdb->opendb_env (b.cstr (), opts)) {
    warn << "cannot open block db\n";
    exit (1);
  }
}

void
ns_impl::exit_cleanup ()
{
  if (_bdb)
    _bdb->closedb ();
  _bdb = 0;
  _dhc = 0;
}

ns_impl::~ns_impl ()
{
  exit_cleanup ();
  if (ns_debug)
    warn << "downloaded " << bytes_downloaded << " bytes\n";
}

bigint
ns::pkid (ptr<sfspriv> sk)
{
  // Return the ChordID that corresponds to the public key of <sk>
  // This is just a calculation.  Perhaps it should be moved to a 
  // utility section instead of in ns.C
  sfs_pubkey2 pk;
  sk->export_pubkey (&pk);
  return keyhash_payload ().id (pk);
}

static void
retrieve_return (ptr<ns_block> b, cbblock cb)
{
  cb (b);
}

//
// local cache
//

ptr<ns_block>
ns_impl::retrieve_local (bigint n)
{
  ptr<dbrec> v = _bdb->lookup (bigint2dbrec (n));
  if (v) {
    str x = dbrec2str (v);
    ptr<ns_block> c = New refcounted<ns_block> (x.cstr ()+20, x.len ()-20);
    return c;
  }
  return 0;
}

bool
ns_impl::check_local (bigint n)
{
  ptr<dbrec> v = _bdb->lookup (bigint2dbrec (n));
  if (v)
    return true;
  return false;
}

bigint
ns_impl::insert_local (const char *data, size_t len)
{
  ptr<dbrec> v = New refcounted<dbrec> (data, 20+len, 20);
  mpz_get_rawmag_be (v->value, 20, &_project);
  bigint n = compute_sha1_hash (v->value, v->len);

  ptr<dbrec> k = bigint2dbrec (n);
  if (_bdb->lookup (k))
    _bdb->remove (k);
  _bdb->insert (k, v);
  return n;
}

//
// flow control for content-hash blocks over the net
//

void
ns_impl::next ()
{
  delaycb (0, wrap (this, &ns_impl::run_next));
}

void
ns_impl::insert_cb (ptr<bool> exp, cbinsertgw_t cb,
                    dhash_stat status, ptr<insert_info> i)
{
  if (*exp)
    return;
  *exp = true;
  _fctr--;
  cb (status, i);
  next ();
}

void
ns_impl::insert_timedout (ptr<bool> exp, bigint n, cbinsertgw_t cb)
{
  if (*exp)
    return;
  *exp = true;
  rpc_timed_out ();
  ptr<insert_info> i = New refcounted<insert_info> (n, vec<bigint> ());
  cb (DHASH_TIMEDOUT, i);
}

void
ns_impl::retrieve_cb (ptr<bool> exp, cb_cret cb,
                      dhash_stat status, ptr<dhash_block> b, vec<chordID> rt)
{
  if (*exp)
    return;
  *exp = true;
  _fctr--;
  if (b)
    bytes_downloaded += b->len;
  cb (status, b, rt);
  next ();
}

void
ns_impl::retrieve_timedout (ptr<bool> exp, cb_cret cb)
{
  if (*exp)
    return;
  *exp = true;
  rpc_timed_out ();
  cb (DHASH_TIMEDOUT, (ptr<dhash_block>) 0L, vec<chordID> ());
}

void
ns_impl::run_next ()
{
  while (_fctr < rpc_window && _isinsert.size () > 0) {
    _fctr++;
    bool i = _isinsert.pop_front ();
    if (i) {
      bigint n = _ibid.pop_front ();
      dhash_ctype ct = _itype.pop_front ();
      ptr<ns_block> b = _iblock.pop_front ();
      ptr<option_block> opt = _iopt.pop_front ();
      cbinsertgw_t cb = _icbs.pop_front ();

      if (_connected) {
        ptr<bool> exp = New refcounted<bool>;
        *exp = false;
        delaycb (rpc_timer, wrap (this, &ns_impl::insert_timedout, exp, n, cb));
        _dhc->insert
	  (n, b->data, b->len, wrap (this, &ns_impl::insert_cb, exp, cb),
	   _project, opt, ct);
      }
      else {
        ptr<insert_info> i = New refcounted<insert_info> (n, vec<bigint> ());
	cb (DHASH_TIMEDOUT, i);
      }
    }
    else {
      bigint n = _rbid.pop_front ();
      dhash_ctype ct = _rtype.pop_front ();
      ptr<option_block> opt = _ropt.pop_front ();
      cb_cret cb = _rcbs.pop_front ();
      if (_connected) {
        ptr<bool> exp = New refcounted<bool>;
        *exp = false;
        delaycb (rpc_timer, wrap (this, &ns_impl::retrieve_timedout, exp, cb));
        _dhc->retrieve
	  (n, ct, wrap (this, &ns_impl::retrieve_cb, exp, cb), _project, opt);
      }
      else
        cb (DHASH_TIMEDOUT, (ptr<dhash_block>) 0L, vec<chordID> ());
    }
  }
}

//
// content-hash block retrieve
//

void
ns_impl::retrieve_block_cb
  (cbblock cb, bigint n, bool skipc,
   dhash_stat status, ptr<dhash_block> b, vec<chordID> rt)
{
  if (status || !b) {
    ptr<ns_block> nb = 0;
    cb (nb);
  }
  else {
    if (!skipc) {
      ptr<dbrec> k = bigint2dbrec (n);
      ptr<dbrec> v = New refcounted<dbrec> (b->data, b->len);
      if (_bdb->lookup (k))
        _bdb->remove (k);
      _bdb->insert (k, v);
    }
    ptr<ns_block> nb = New refcounted<ns_block> (b->data+20, b->len-20);
    cb (nb);
  }
}

void
ns_impl::retrieve_block (bigint n, cbblock cb, bool nocache)
{
  if (!nocache) {
    ptr<dbrec> v = _bdb->lookup (bigint2dbrec (n));
    if (v) {
      str x = dbrec2str (v);
      ptr<ns_block> c = New refcounted<ns_block> (x.cstr ()+20, x.len ()-20);
      delaycb (0, wrap (retrieve_return, c, cb));
      return;
    }
  }
  _rbid.push_back (n);
  _rtype.push_back (DHASH_CONTENTHASH);
  _ropt.push_back (0L);
  _rcbs.push_back (wrap (this, &ns_impl::retrieve_block_cb, cb, n, nocache));
  _isinsert.push_back (false);
  next ();
}

//
// content-hash block insert
//

void
ns_impl::insert_block_cb
  (bool nocache, cbinsert cb, dhash_stat status, ptr<insert_info> i)
{
  if (status != DHASH_OK) {
    warn << "dht insert failed (" << (int) status << "): "
         << armor64A (i->key) << ", sync later\n";
    if (nocache) {
      cb (NS_NOENT, i->key);
      return;
    }
  }
  cb (NS_OK, i->key);
}

bigint
ns_impl::insert_block (const char *data, size_t len, cbinsert cb, bool nocache)
{
  ptr<dbrec> v = New refcounted<dbrec> (data, 20+len, 20);
  mpz_get_rawmag_be (v->value, 20, &_project);
  bigint n = compute_sha1_hash (v->value, v->len);
  ptr<ns_block> b = New refcounted<ns_block> (v->value, v->len);
  
  if (!nocache) {
    ptr<dbrec> k = bigint2dbrec (n);
    if (_bdb->lookup (k))
      _bdb->remove (k);
    _bdb->insert (k, v);
  }

  _ibid.push_back (n);
  _itype.push_back (DHASH_CONTENTHASH);
  _iblock.push_back (b);
  _iopt.push_back (0L);
  _icbs.push_back (wrap (this, &ns_impl::insert_block_cb, nocache, cb));
  _isinsert.push_back (true);
  next ();
  return n;
}

//
// public key insert
//

#define update_pkv(n, version) \
  _pkv.remove (n); \
  _pkv.insert (n, version);

void
ns_impl::insert_pk_cb (ptr<bool> exp, ptr<pk_insert_state> st,
                       dhash_stat status, ptr<insert_info> i)
{
  if (exp) {
    if (*exp)
      return;
    *exp = true;
  }
  assert (i);
  ns_stat s = NS_OK;
  if (status != DHASH_OK)
    warn << "dht insert failed (" << (int) status << "): "
         << armor64A (i->key) << ", sync later\n";
  st->cb (s, i->key);
}

void
ns_impl::insert_pk_timedout (ptr<bool> exp, bigint n, ptr<pk_insert_state> st)
{
  if (*exp)
    return;
  *exp = true;
  rpc_timed_out ();
  ptr<insert_info> i = New refcounted<insert_info> (n, vec<bigint> ());
  insert_pk_cb ((ptr<bool>)0L, st, DHASH_TIMEDOUT, i);
}

extern int
pk_put_block_contents
(sfs_pubkey2 &key, sfs_sig2 &sig, keyhash_payload& p, xdrsuio &x);

void
ns_impl::insert_pk (ptr<pk_insert_state> st)
{
  sfs_pubkey2 pk;
  sfs_sig2 s;
  keyhash_payload p (_project, st->version, st->data);
  p.sign (st->sk, pk, s);

  xdrsuio x;
  if (pk_put_block_contents (pk, s, p, x) != 0) {
    warn << "cannot encode pk block\n";
    ptr<insert_info> i = New refcounted<insert_info>
      (p.id (pk), vec<bigint> ());
    delaycb (0, wrap (this, &ns_impl::insert_pk_cb,
	              (ptr<bool>)0L, st, DHASH_TIMEDOUT, i));
    return;
  }

  int m_len = x.uio ()->resid ();
  char *m_dat = suio_flatten (x.uio ());
  ptr<dbrec> v = New refcounted <dbrec> (m_dat, m_len);
 
  bigint n = p.id (pk);
  ptr<dbrec> k = bigint2dbrec (n);
  if (_bdb->lookup (k))
    _bdb->remove (k);
  _bdb->insert (k, v);
  update_pkv (n, st->version);

  bool c = _connected;
  for (unsigned i=0; i<_stalepks.size () && c; i++)
    if (_stalepks [i] == n)
      c = false;

  if (c) {
    ptr<bool> exp = New refcounted<bool>;
    *exp = false;
    delaycb (rpc_timer, wrap (this, &ns_impl::insert_pk_timedout, exp, n, st));
    _dhc->insert
      (n, pk, s, p, wrap (this, &ns_impl::insert_pk_cb, exp, st), _project, 0L);
  }
  else {
    ptr<insert_info> i = New refcounted<insert_info> (n, vec<bigint> ());
    delaycb (0, wrap (this, &ns_impl::insert_pk_cb,
	              (ptr<bool>)0L, st, DHASH_TIMEDOUT, i));
  }
}

void
ns_impl::fetch_pk_for_insert
  (bigint n, ptr<pk_insert_state> st, ptr<keyhash_payload>)
{
  long v = 0;
  if (_pkv [n])
    v = (*(_pkv [n]))+1;
  else if (ns_debug)
    warn << "cannot fetch " << armor64A (n) << ", set version to 0\n";
  st->version = v;
  insert_pk (st);
}

void
ns_impl::insert_pk (ptr<sfspriv> sk, str data, cbinsert cb)
{
  ptr<pk_insert_state> st = New refcounted<pk_insert_state> (sk, data, cb);
  bigint hash = ns::pkid (st->sk);
  long *n = _pkv [hash];
  if (n) {
    st->version = (*n)+1;
    insert_pk (st);
  }
  else
    retrieve_pk (hash, wrap (this, &ns_impl::fetch_pk_for_insert, hash, st));
}

void
insert_raw_cb (cbinsert cb, bigint n, dhash_stat status, ptr<insert_info> i)
{
  if (status)
    cb (NS_ERR, n);
  else
    cb (NS_OK, n);
}

void
ns_impl::insert_raw_pk_from_cache (bigint n, cbinsert cb)
{
  if (!_connected) {
    cb (NS_ERR, n);
    return;
  }
  ptr<dbrec> v = _bdb->lookup (bigint2dbrec (n));
  if (!v) {
    cb (NS_NOENT, n);
    return;
  }

  ptr<dhash_block> b =
    New refcounted<dhash_block> (v->value, v->len, DHASH_KEYHASH);
  ptr<dhash_block> bb = get_block_contents (b, DHASH_KEYHASH);
  ptr<keyhash_payload> p = keyhash_payload::decode (bb);

  if (!p) {
    cb (NS_NOENT, n);
    return;
  }

  xdrmem x (b->data, b->len, XDR_DECODE);
  sfs_pubkey2 k;
  sfs_sig2 s;
  if (!xdr_sfs_pubkey2 (&x, &k) ||
      !xdr_sfs_sig2 (&x, &s)) {
    cb (NS_NOENT, n);
    return;
  }
  
  _dhc->insert
    (p->id (k), k, s, *p, wrap (insert_raw_cb, cb, p->id (k)), _project, 0L);
}

//
// public key retrieve
//

void 
ns_impl::retrieve_pk_cb
  (bigint n, ptr<dhash_block> cached, cbkeyhash cb,
   dhash_stat status, ptr<dhash_block> b, vec<chordID> rt)
{
  ptr<keyhash_payload> p = 0;
  ptr<keyhash_payload> cached_p;
  ptr<dhash_block> cached_b = 0;

  if (status == DHASH_SAMEVERSION && cached) {
    if (ns_debug)
      warn << "pk on server is the same as cached, no need to fetch\n";
    status = DHASH_OK;
    b = cached;
  }
  if (status || !b) {
    b = 0;
    p = 0;
  }
  else if (b) {
    ptr<dhash_block> bb = get_block_contents (b, DHASH_KEYHASH);
    if (!bb || !(p = keyhash_payload::decode (bb))) {
      if (bb)
	warn << "had block, could not decode\n";

      b = 0;
      p = 0;
    }
  }

  if (cached &&
      (cached_b = get_block_contents (cached, DHASH_KEYHASH)) &&
      (cached_p = keyhash_payload::decode (cached_b))) {
    if (!p || cached_p->version () > p->version ()) {
      warn << "cached pubkey block is more recent, no need to fetch\n";
      if (ns_debug) {
	warn << "cached version " << cached_p->version () << "\n";
	if (p)
	  warn << "dht version " << p->version () << "\n";
      }
      b = cached;
      p = cached_p;
      _stalepks.push_back (n);
    }
  }
  else if (!p && ns_debug) {
    if (!cached)
      warn << "no cached block\n";
    else if (!cached_b)
      warn << "no cached content\n";
    else if (!cached_p)
      warn << "no cached payload\n";
    warn << "no cached copy of " << armor64A (n) << "\n";
  }

  if (p == 0) {
    cb (p);
    return;
  }

  if (b) {
    ptr<dbrec> v = New refcounted<dbrec> (b->data, b->len);
    ptr<dbrec> k = bigint2dbrec (n);
    if (_bdb->lookup (k))
      _bdb->remove (k);
    _bdb->insert (k, v);
  }

  long *x = _pkv [n];
  if (x && (*x) > p->version ()) {
    warn << "stale record v" << p->version ()
         << ", should probably be v" << (*x) << "\n";
    p = 0;
  }
  else
    update_pkv (n, p->version ());
  cb (p);
}

void
ns_impl::retrieve_pk_timedout (ptr<bool> exp, cb_cret cb)
{
  if (exp) {
    if (*exp)
      return;
    *exp = true;
    rpc_timed_out ();
  }
  cb (DHASH_TIMEDOUT, (ptr<dhash_block>) 0L, vec<chordID> ());
}

void
ns_impl::retrieve_pk_ret
  (ptr<bool> exp, cb_cret cb,
   dhash_stat status, ptr<dhash_block> b, vec<chordID> rt)
{
  if (*exp)
    return;
  *exp = true;
  if (b)
    bytes_downloaded += b->len;
  cb (status, b, rt);
}

void
ns_impl::retrieve_pk (bigint n, cbkeyhash cb)
{
  ptr<dhash_block> b = 0L;
  ptr<dbrec> v = _bdb->lookup (bigint2dbrec (n));
  if (!v) {
    if (ns_debug)
      warn << "cache miss: " << armor64A (n) << "\n";
  }
  else
    b = New refcounted<dhash_block> (v->value, v->len, DHASH_KEYHASH);

  ptr<keyhash_payload> cp = 0;
  ptr<dhash_block> bb = 0L;
  if (b) {
    bb = get_block_contents (b, DHASH_KEYHASH);
    if (!bb) {
      warn << "cannot get block contents!\n";
      b = 0;
    }
    else
      cp = keyhash_payload::decode (bb);
    if (!cp) {
      warn << "cannot decode payload!\n";
      b = 0;
    }
  }

  ptr<bool> exp = New refcounted<bool>;
  *exp = false;
  delaycb (rpc_timer, wrap (this, &ns_impl::retrieve_pk_timedout, exp,
	                    wrap (this, &ns_impl::retrieve_pk_cb, n, b, cb)));

  if (_connected && b) {
    ptr<option_block> opt = New refcounted<option_block>;
    opt->flags = DHASHCLIENT_RAW;
    _dhc->retrieve_pk
      (n, wrap (this, &ns_impl::retrieve_pk_ret,
		exp, wrap (this, &ns_impl::retrieve_pk_cb, n, b, cb)),
       cp->version (), _project, opt);
  }
  else if (_connected) {
    ptr<option_block> opt = New refcounted<option_block>;
    opt->flags = DHASHCLIENT_RAW;
    _dhc->retrieve
      (n, DHASH_KEYHASH,
       wrap (this, &ns_impl::retrieve_pk_ret, exp,
	     wrap (this, &ns_impl::retrieve_pk_cb, n, b, cb)), _project, opt);
  }
  else
    delaycb
      (0, wrap (this, &ns_impl::retrieve_pk_timedout,
		(ptr<bool>)0L,
		wrap (this, &ns_impl::retrieve_pk_cb, n, b, cb)));
}

void
ns_impl::retrieve_pk_nc_cb
  (bigint n, cbkeyhash cb,
   dhash_stat status, ptr<dhash_block> b, vec<chordID> rt)
{
  ptr<keyhash_payload> p = 0;
  if (!status && b) {
    ptr<dhash_block> bb = get_block_contents (b, DHASH_KEYHASH);
    if (bb)
      p = keyhash_payload::decode (bb);
  }
  cb (p);
}

void
ns_impl::retrieve_pk_from_net (bigint n, cbkeyhash cb)
{
  if (!_connected) {
    cb ((ptr<keyhash_payload>) 0L);
    return;
  }
  ptr<bool> exp = New refcounted<bool>;
  *exp = false;
  delaycb
    (rpc_timer, wrap (this, &ns_impl::retrieve_pk_timedout, exp,
		      wrap (this, &ns_impl::retrieve_pk_nc_cb, n, cb)));

  ptr<option_block> opt = New refcounted<option_block>;
  opt->flags = DHASHCLIENT_RAW;
  _dhc->retrieve
    (n, DHASH_KEYHASH,
     wrap (this, &ns_impl::retrieve_pk_ret, exp,
           wrap (this, &ns_impl::retrieve_pk_nc_cb, n, cb)), _project, opt);
}

void
ns_impl::checkpks_cb (ptr<bool> exp, cbcheckpks cb,
                      dhash_stat status, vec<bigint> n, vec<long> v)
{
  if (*exp)
    return;
  *exp = true;
  if (status)
    cb (n, NS_ERR);
  else
    cb (n, NS_OK);
}

void
ns_impl::check_pks_timedout (ptr<bool> exp, cbcheckpks cb)
{
  if (*exp)
    return;
  *exp = true;
  rpc_timed_out ();
  cb (vec<bigint> (), NS_ERR);
}

static void
checkpks_err_delayed (cbcheckpks cb)
{
  cb (vec<bigint> (), NS_ERR);
}

void
ns_impl::checkpks (vec<bigint> n, vec<long> v, cbcheckpks cb)
{
  assert (n.size () == v.size ());
  if (_connected) {
    ptr<bool> exp = New refcounted<bool>;
    *exp = false;
    delaycb (rpc_timer, wrap (this, &ns_impl::check_pks_timedout, exp, cb));
    _dhc->check_pks
      (_project, n, v, wrap (this, &ns_impl::checkpks_cb, exp, cb));
  }
  else
    delaycb (0, wrap (checkpks_err_delayed, cb));
}

//
// wrappers
//

static void
retrieve_commit_cb (cbcrec cb, ptr<ns_block> b)
{
  if (b) {
    crec c = commit_record::alloc (str (b->data, b->len));
    cb (c, c ? NS_OK : NS_NOENT);
  }
  else
    cb ((crec) 0, NS_NOENT);
}

void
ns_impl::retrieve_commit (bigint n, cbcrec cb)
{
  retrieve_block (n, wrap (retrieve_commit_cb, cb));
}

static void 
retrieve_lptr_cb (bigint n, cblptr cb, ptr<keyhash_payload> p)
{
  lptr u = 0;
  long v = 0;
  if (p && !(u = log_pointer::alloc (p->buf ()))) {
    warn << armor64A (n) << ": not a lptr block\n";
    u = 0;
  }
  else if (p && u)
    v = p->version ();

  cb (u, v, u ? NS_OK : NS_NOENT);
}

void
ns_impl::retrieve_lptr (bigint n, cblptr cb)
{
  retrieve_pk (n, wrap (retrieve_lptr_cb, n, cb));
}

