#include "dhash_common.h"
#include "ptr_block.h"

static int ptrblock_debug = 0;

ptr_block::ptr_block() :
_revision(0)
{
}

ptr_block::ptr_block(const ptr_block &original) {
  for(unsigned i=0; i<original._pointers.size(); i++) {
    _pointers.push_back(original._pointers[i]);
    _labels.push_back(original._labels[i]);
  }
  sort();
}

void
ptr_block::read_block (bigint ptr_block_addr, callback<void, log_status>::ref cb) {
  nsobj->retrieve_pk(ptr_block_addr, wrap(this, &ptr_block::got_ptr_block, cb));
}

ptr_block::~ptr_block() {}

void
ptr_block::got_ptr_block (callback<void, log_status>::ref cb,
			  ptr<keyhash_payload> p)
{
  if (p == 0) {
    warn << "Could not fetch pointer block.\n";
    cb (LOGERR_NOBLOCK);
    return;
  }

  if (ptrblock_debug) {
    u_int64_t t = getusec () - _fetcht;
    warn << "fetchproj: reply, +" << (long) t << "\n";
  }

  _revision = p->version();
  xin in (p->buf ());
  if (!parse (&in)) {
    cb (LOGERR_BADPARSE);
    return;
  }
  cb (LOG_OK);
}

bool ptr_block::add_entry (bigint pointer, str label) {
  // This will not add the entry if the this already
  // contains an entry with the same label.  However,
  // it will overwrite an existing label if this contains
  // a entry for <pointer>.

  for(unsigned i=0; i<_labels.size(); i++)
    if (strncasecmp (_labels[i].cstr (), label, label.len ()) == 0 &&
	_pointers[i] != pointer) {
      return false;
    }
  for(unsigned i=0; i<_pointers.size(); i++) {
    if (pointer == _pointers[i]) {
      _labels[i] = label;
      return true;
    }
  }
  _pointers.push_back(pointer);
  _labels.push_back(label);
  sort();
  return true;
}

bigint 
ptr_block::store (ptr<sfspriv> sk, callback<void, ns_stat, bigint>::ref cb) {
  str data = serialize ();
  bigint n = ns::pkid (sk);
  nsobj->insert_pk (sk, data, cb);
  return n;
}

bigint
ptr_block::get_ptr(unsigned i) const {
  return _pointers[i];
}

str 
ptr_block::get_label(unsigned i) const {
  return _labels[i];
}

unsigned 
ptr_block::nentries() const {
  return _pointers.size();
}

str
ptr_block::serialize() {
  sort (); // paranoia
  strbuf sb;
  sstr (sb, VERSION);
  sint (sb, _pointers.size ());
  for (unsigned i=0; i<_pointers.size (); i++) {
    sbigint (sb, _pointers[i]) ;
    sbigint (sb, 0);
    sstr (sb, _labels[i]);
  }
  return sb;
}

bool
ptr_block::parse (xin *in)
{
  str swver;
  int n;
  if (!in->rstr (swver))
    return false;
  if (!in->rint (n))
    return false;

  for (int i=0; i<n; i++) {
    bigint p, aux;
    str l;
    if (!in->rbigint (p) ||
	!in->rbigint (aux) ||
	!in->rstr (l))
      return false;
    if (!l || l == "")
      warn << "Pointer " << armor64A (p) << " has no label\n";
    add_entry (p, l);
  }

  sort ();
  return true;
}

void
ptr_block::sort()
  // slow but simple sort
{
  vec<bigint> h;
  vec<str> n;
  vec<bool> done;
  bool found = true;
  for (unsigned i=0; i<_pointers.size(); i++)
    done.push_back(false);

  while (found) {
    found = false;
    bigint sv;
    unsigned s = 0;
    for (unsigned i=0; i<_pointers.size(); i++) {
      if (!done[i]) {
	if (!found || _pointers[i] < sv) {
          found = true;
	  sv = _pointers[i];
	  s = i;
	}
      }
    }
    if (found) {
      h.push_back(_pointers[s]);
      n.push_back(_labels[s]);
      done[s] = true;
    }
  }

  _pointers = h;
  _labels = n;
}

long
ptr_block::get_blk_revision () const
{
  return _revision;
}

void
ptr_block::print(str prefix) const
{
  for(unsigned i=0; i<_pointers.size(); i++)
    warn << prefix << _labels[i] << " (" << armor64A(_pointers[i]) << ")\n";
}
