/*
 *  Copyright (C) 2002-2004  Massachusetts Institute of Technology
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License as
 *  published by the Free Software Foundation; either version 2, or (at
 *  your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 *  USA
 */

#include "ns.h"
#include "pack.h"
#include "stdinc.h"
#include "record.h"

log_pointer::log_pointer ()
{
  _lock = 0;
  _verf = 0;
  _commit = 0;
}

ptr<log_pointer>
log_pointer::alloc (str s)
{
  xin in (s);
  ptr<log_pointer> x = New refcounted<log_pointer> ();

  str swver;
  if (!in.rstr (swver))
    return 0;

  bigint n;
  if (!in.rbigint (x->_commit) ||
      !in.rbigint (x->_lock) ||
      !in.rbigint (x->_verf) ||
      !in.ruint (x->_pw_seqn) ||
      !in.rbigint (n))
    return 0;

  return x;
}

str
log_pointer::serialize () const
{
  strbuf sb;
  bigint n = 0;
  sstr (sb, VERSION);
  sbigint (sb, _commit);
  sbigint (sb, _lock);
  sbigint (sb, _verf);
  suint (sb, _pw_seqn);
  sbigint (sb, n);
  return sb;
}

#define NHASHES 256

commit_record::commit_record ()
  : _nhashes (NHASHES)
{
  _head = 0;
  _btag = "";
  _proposed_name = "";
  _msg = "";
  _ndeltas = 0;
  struct timeval t;
  gettimeofday (&t, 0);
  _time = t.tv_sec;
  _flags = 0;
  _supersede = 0;
}

commit_record::commit_record (const commit_info& c, str n, str m, str btag,
                              unsigned flags, bigint supersede, long commit_time)
  : _nhashes (NHASHES)
{
  _head = 0;
  _depends = c;
  _depends.vptr = 0;
  assert (n);
  assert (m);
  _proposed_name = n;
  _msg = m;

  _ndeltas = 0;
  if (commit_time) {
    _time = commit_time;
  } else {
    struct timeval t;
    gettimeofday (&t, 0);
    _time = t.tv_sec;
  }

  if (btag)
    _btag = btag;
  else
    _btag = "";
  _flags = flags;
  _supersede = supersede;
}

str
commit_record::serialize () const
{
  strbuf sb;
  sstr (sb, VERSION);
  sbigint (sb, _head);
  sbigint (sb, _prev);
  suint (sb, _time);
  _depends.serialize (sb);
  sstr (sb, _btag);

  assert (_nhashes == NHASHES);
  suint (sb, _nhashes);
  assert (_ndeltas >= _deltas.size ());
  suint (sb, _ndeltas);
  suint (sb, _deltas.size ());
  for (unsigned i=0; i<_deltas.size (); i++)
    sbigint (sb, _deltas [i]);
  sbigint (sb, _sind);
  sbigint (sb, _dind);

  sstr (sb, _proposed_name);
  sstr (sb, _msg);
  suint (sb, _flags);
  sbigint (sb, _supersede);
  return sb;
}

ptr<commit_record>
commit_record::alloc (str s)
{
  xin in (s);
  unsigned n;
  ptr<commit_record> x = New refcounted<commit_record> ();

  if (!in.rstr (x->_swver))
    return 0;

  if (!in.rbigint (x->_head) ||
      !in.rbigint (x->_prev) ||
      !in.ruint (x->_time) ||
      !commit_info::parse (&in, x->_depends))
    return 0;

  if (!in.rstr (x->_btag) ||
      !in.ruint (x->_nhashes) ||
      !in.ruint (x->_ndeltas) ||
      !in.ruint (n))
    return 0;

  for (unsigned i=0; i<n; i++) {
    bigint r;
    if (!in.rbigint (r))
      return 0;
    x->_deltas.push_back (r);
  }

  if (!in.rbigint (x->_sind) ||
      !in.rbigint (x->_dind) ||
      !in.rstr (x->_proposed_name) ||
      !in.rstr (x->_msg) ||
      !in.ruint (x->_flags) ||
      !in.rbigint (x->_supersede))
    return 0;

  return x;
}

bool
delta::parse_meta (xin *in)
{
  str swver;
  if (!in->rstr (swver)) {
    warn << "cannot parse version!\n";
    return false;
  }

  if (!in->ruint (_rtype)) {
    warn << "cannot parse metadata\n";
    return false;
  }
  return true;
}

bool
delta::parse_type (xin *in, unsigned &type)
{
  str swver;
  if (!in->rstr (swver)) {
    warn << "cannot parse version!\n";
    return false;
  }

  if (!in->ruint (type))
    return false;
  return true;
}


static void
get_sind (unsigned ndeltas, callback<void,unsigned>::ref xcb, crec c,
          callback<void,bigint,bool>::ref cb, ptr<ns_block> b)
{
  if (b) {
    str s (b->data, b->len);
    xin x (s);
    for (unsigned i=0; i<c->nhashes () && ndeltas > 0; i++) {
      ndeltas--;
      bigint n;
      if (x.rbigint (n))
	cb (n, false);
      else {
	cb ((bigint) 0, true);
	return;
      }
    }
    xcb (ndeltas);
  }
  else
    cb ((bigint) 0, true);
}

static void
next_sind (crec c, callback<void,bigint>::ref ind_cb,
           callback<void,bigint,bool>::ref cb,
           ptr<xin> x, ptr<str> s, unsigned ndeltas)
{
  if (ndeltas > 0) {
    bigint n;
    if (x->rbigint (n)) {
      ind_cb (n);
      nsobj->retrieve_block
        (n, wrap (get_sind, ndeltas,
		  wrap (next_sind, c, ind_cb, cb, x, s), c, cb));
    }
    else
      cb ((bigint) 0, true);
  }
}

static void
get_dind (unsigned ndeltas, crec c, callback<void,bigint>::ref ind_cb,
          callback<void,bigint,bool>::ref cb, ptr<ns_block> b)
{
  if (b) {
    ptr<str> s = New refcounted<str> (b->data, b->len);
    ptr<xin> x = New refcounted<xin> (*s);
    bigint n;
    if (x->rbigint (n)) {
      ind_cb (n);
      nsobj->retrieve_block
        (n, wrap (get_sind, ndeltas,
		  wrap (next_sind, c, ind_cb, cb, x, s), c, cb));
    }
    else
      cb ((bigint) 0, true);
  }
  else
    cb ((bigint) 0, true);
}

static void
check_dind (crec c, callback<void,bigint>::ref ind_cb,
            callback<void,bigint,bool>::ref cb, unsigned ndeltas)
{
  if (ndeltas > 0) {
    bigint d = c->dind ();
    ind_cb (d);
    nsobj->retrieve_block (d, wrap (get_dind, ndeltas, c, ind_cb, cb));
  }
}

void
get_deltas (crec c, callback<void, bigint>::ref ind_cb,
                 callback<void, bigint, bool>::ref record_cb)
{
  unsigned ndeltas = c->ndeltas ();
 
  // first, return direct pointers
  const vec<bigint>& v = c->deltas ();
  for (unsigned i=0; i<c->nhashes () && ndeltas > 0; i++) {
    ndeltas--;
    record_cb (v [i], false);
  }

  if (ndeltas > 0) { // single indirect
    bigint s = c->sind ();
    ind_cb (s);
    nsobj->retrieve_block
      (s, wrap (get_sind, ndeltas, wrap (check_dind, c, ind_cb, record_cb),
		c, record_cb));
  }
}

