#include "gcc_compat.h"
#include "merkle_server.h"
#include <location.h>
#include <locationtable.h>
#include <comm.h>
#include "transport_prot.h"
#include "blacklist.h"

// ---------------------------------------------------------------------------
// merkle_server

merkle_server::merkle_server (ptr<dhash_blockstore> storage, addHandler_t addHandler, 
			      missingfnc2_t missingfnc, vnode *host_node)
  : storage (storage), host_node (host_node), missingfnc (missingfnc)
{
  (*addHandler) (merklesync_program_1, wrap(this, &merkle_server::dispatch));
}


void
merkle_server::missing (ptr<location> n, bigint key)
{
  (*missingfnc) (n, key);
}

void
merkle_server::doRPC (ptr<location> dst, RPC_delay_args *args)
{
  host_node->doRPC (dst, args->prog, args->procno, args->in, 
		    args->out, args->cb);
}

void
merkle_server::dispatch (user_args *sbp)
{
  if (!sbp)
    return;

  switch (sbp->procno) {
  case MERKLESYNC_SENDNODE:
    // request a node of the merkle tree
    {
      // Get remote sides ip:port and chordID
      chord_node from;
      sbp->fill_from (&from);
      ptr<location> rloc = host_node->locations->lookup_or_create (from);
      
      sendnode_arg *arg = sbp->TEMPLATE getarg<sendnode_arg> ();
      merkle_rpc_node *rnode = &arg->node;
      merkle_node *lnode;
      u_int lnode_depth;
      merkle_hash lnode_prefix;

      vec<bigint> exclu_proj;
      blacklist::only().append_blacklist(host_node->my_location()->address(), exclu_proj);
      blacklist::only().append_blacklist(from.r, exclu_proj);
      ptr<merkle_tree> ltree = get_mtree (from.x, arg->rngmin, arg->rngmax, exclu_proj);

      lnode = ltree->lookup (&lnode_depth, rnode->depth, rnode->prefix);
      if (lnode_depth != rnode->depth) {
	warn << "local depth ( " << lnode_depth 
	     << ") is not equal to remote depth (" << rnode->depth << ")\n";
	warn << "prefix is " << rnode->prefix << "\n";
      }
      lnode_prefix = rnode->prefix;
      // XXX isn't this a NOP, given the previous assertion about equal depths 
      lnode_prefix.clear_suffix (lnode_depth);

      compare_nodes (ltree, arg->rngmin, arg->rngmax, lnode, rnode,
		     wrap (this, &merkle_server::missing, rloc),
		     wrap (this, &merkle_server::doRPC, rloc));

      sendnode_res res (MERKLE_OK);
      format_rpcnode (ltree, lnode_depth, lnode_prefix, 
		      lnode, &res.resok->node);
      sbp->reply (&res);
      break;
    }
     
  case MERKLESYNC_GETKEYS:
    {
      getkeys_arg *arg = sbp->TEMPLATE getarg<getkeys_arg> ();

      vec<ptr<dbrec> > keys;
      storage->getchblockids
	(&keys, id2dbrec(arg->rngmin), id2dbrec(arg->rngmax), 65);

      bool more = false;
      if (keys.size () == 65) {
	more = true;
	keys.pop_back();
      }

      getkeys_res res (MERKLE_OK);
      res.resok->morekeys = more;
      res.resok->keys.setsize (keys.size ());
      for (u_int i = 0; i < keys.size (); i++)
	res.resok->keys[i] = to_merkle_hash (keys[i]);

      sbp->reply (&res);
      break;
    }

  default:
    fatal << "unknown proc in merkle " << sbp->procno << "\n";
    sbp->reject (PROC_UNAVAIL);
    break;
  }
}

ptr<merkle_tree> 
merkle_server::get_mtree (chordID node_id, chordID min, chordID max, vec<bigint> exclu_proj) {
  int MTREE_TIMEOUT = 10;

  timeval tv;
  mtree_wrap *mwrap = mtree_cache[node_id];
  
  if (!mwrap) {
    ptr<merkle_tree> mtree = 
      New refcounted<merkle_tree> (storage, min, max, exclu_proj);
    mwrap = New mtree_wrap(node_id, mtree);
    mtree_cache.insert(mwrap);
  }
  
  assert (mwrap = mtree_cache[node_id]);
  gettimeofday(&tv, NULL);
  
  if (tv.tv_sec - mwrap->update_time.tv_sec > MTREE_TIMEOUT)
    mwrap->update_mtree( New refcounted<merkle_tree> (storage, min, max, exclu_proj) );

  return mwrap->tree;
}
