/*
 *  Copyright (C) 2002-2004  Massachusetts Institute of Technology
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License as
 *  published by the Free Software Foundation; either version 2, or (at
 *  your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 *  USA
 */

#include <sys/types.h>
#include <dirent.h>
#include <fnmatch.h>
#include "stdinc.h"
#include "pastwatch.h"
#include "rxx.h"

typedef callback<void, bigint, int>::ref fhcb_t;
typedef callback<void, const pde_t &, fhcb_t >::ref recursecb_t;
typedef ptr<vec<str> > svp;

static vec<str> ifds;
static vec<svp> ifs;

struct recurse_entry {
  pde_t pde;
  callback<void, bigint, int>::ptr post_cb;
  bool dir;

  recurse_entry() :
    post_cb( NULL )
  { }
  recurse_entry(pde_t e, fhcb_t cb) :
    post_cb(cb)
  {
    pde.name = e.name;
    pde.path = e.path;
    pde.cfspath = e.cfspath;
    pde.type = e.type;
    pde.mode = e.mode;
    pde.fh = e.fh;
    pde.dirfh = e.dirfh;
  }
};

static svp
get_ignored_files (str dir)
{
  for (unsigned i=0; i<ifds.size (); i++)
    if (ifds [i] == dir)
      return ifs [i];

  svp s = New refcounted<vec<str> >;
  s->clear ();
  str p = dir << "/.pastwatchignore";
  str buf = file2str (p);
  if (buf) {
    vec<str> n;
    split (&n, "\n", buf);
    for (unsigned i=0; i<n.size (); i++)
      s->push_back (n [i]);
  }
  ifds.push_back (dir);
  ifs.push_back (s);
  return s;
}

void
check_ignore (str dirpath, str name, bool& ignore)
{
  svp xp = get_ignored_files (dirpath);
  for (unsigned i=0; i<xp->size () && !ignore; i++)
    if (fnmatch ((*xp) [i], name, 0) == 0)
      ignore = true;
}

void
check_skip (str dirpath, str name, bool& skip)
{
  if (strcmp (name, ".") == 0 || strcmp (name, "..") == 0)
    skip = true;
  if (name == "/" || strcmp (name.cstr (), PASTWATCH_REF) == 0)
    skip = true;
}

int
check_dirent (str dirpath, dirent *e, bool& ignore)
{
  if (e->d_type == 0) {
    str b = dirpath << "/" << e->d_name;
    struct stat sb;
    if (lstat (b, &sb) == 0) {
      if ((sb.st_mode & S_IFLNK) == S_IFLNK)
	e->d_type = DT_LNK;
      else if ((sb.st_mode & S_IFREG) == S_IFREG)
	e->d_type = DT_REG;
      else if ((sb.st_mode & S_IFDIR) == S_IFDIR)
	e->d_type = DT_DIR;
    }
  }
  if (e->d_type != DT_DIR && e->d_type != DT_REG && e->d_type != DT_LNK) {
    warn << "skipping " << e->d_name << ": file type "
         << e->d_type << " not supported\n";
    return -1;
  }

  check_skip (dirpath, str (e->d_name), ignore);
  if (ignore)
    return 0;
  check_ignore (dirpath, str (e->d_name), ignore);
  return 0;
}

class recurse_object : public virtual refcount
{
  int work;   // number of files/dirs that are currently being traversed.
  unsigned max_outstanding;
  unsigned outstanding;
  int error;
  recursecb_t iter_cb;
  cbi done_cb;

  vec< ptr<recurse_entry> > queue;
  
  void enqueue (const pde_t& e, fhcb_t cb) {
    queue.push_back(New refcounted<recurse_entry>(e, cb));
    dequeue();
  }

  void dequeue () {
    while (outstanding < max_outstanding && queue.size()) {
      outstanding++;
      ptr<recurse_entry> entry = queue.pop_front();
      iter_cb (entry->pde, entry->post_cb);
    }
  }

  void recurse_reply (str path, recursecb_t iter_cb, cbi cb, bool dir,
		      bigint fh, int err)
  {
    outstanding--;
    work --;
    dequeue();

    if (err) {
      error = err;
      if (work == 0)
        cb (error);
      return;
    }
    if (dir)
      do_recurse (path, iter_cb, cb, fh);
    if (work == 0)
      cb (error);
  }
  
  void
  do_recurse (str dirpath, recursecb_t iter_cb, cbi cb, bigint dirfh)
  {
    work ++;
    
    DIR *dir;
    dir = opendir (dirpath);
  
    if (dir == 0) {
      work --;
      error = -1;
      perror ("opendir");
      if (work == 0)
        cb (error);
      return;
    }

    struct dirent *e;
    while ((e = readdir (dir)) && !error) {
      bool should_ignore = false;
      if (check_dirent (dirpath, e, should_ignore))
	continue;
      if (!should_ignore) {
        strbuf b = dirpath << "/" << e->d_name;
	pde_t pde;
	pde.name = str (e->d_name);
	pde.path = b;
	if (e->d_type == DT_DIR)
	  pde.type = CFSDIR;
	else if (e->d_type == DT_REG)
	  pde.type = CFSREG;
	else if (e->d_type == DT_LNK)
	  pde.type = CFSLNK;
        pde.dirfh = dirfh;
	work++;

	enqueue (pde, wrap (mkref(this), &recurse_object::recurse_reply, 
			    pde.path, iter_cb, cb, e->d_type == DT_DIR));
      }
    }
  
    closedir (dir);
    work --;
    if (work == 0)
      cb (error);
  }

public:
  recurse_object (str dirpath, recursecb_t _iter_cb, cbi cb, bigint dirfh, unsigned _max_outstanding) : 
    work (0), max_outstanding(_max_outstanding), 
    outstanding(0), error (0), iter_cb(_iter_cb), done_cb (cb)
  {
    do_recurse (dirpath, _iter_cb, cb, dirfh);
  }

  ~recurse_object() {
  }
};

void
recurse (str dirpath, recursecb_t iter_cb, cbi cb, bigint dirfh, unsigned max_outstanding)
{
  ptr<recurse_object> a =
    New refcounted<recurse_object> (dirpath, iter_cb, cb, dirfh, max_outstanding);
}

