#!/usr/bin/perl -w
#
# 2) Should this convert all modules in a repository ?
# 4) Should check to make sure symbolic tags are accurate (They
#    can be applied to a single commit record.
# 5) Also, some files may not be tagged, because they did not exist
#    when the tag was created.  This is ok.  We'll just tag them all
#    inclusively.  Even if only one file was tagged, we'll translate the
#    tag.  
#
# Since CVS is based on files, and Pastwatch is based on the entire
# repository, some properties of a CVS repository cannot be directly
# translated into Pastwatch properties.  Individual file tags are one
# example.  Only an entire repository version can have a symbolic tag.

use Date::Calc (Add_Delta_YMDHMS);
use Date::Parse;

my $pastwatch_root = $ENV{PASTWATCH_ROOT};
if (!$pastwatch_root || $pastwatch_root eq "") {
    my $home = $ENV{HOME};
    if (!$home || $home eq "") {
	die "please define either the PASTWATCH_ROOT or HOME environment variables.";
    }
    $pastwatch_root = "$home/.pastwatch";
}

my $cvsroot = "";
my $cvs_module = "";

my $past_module = "";
my $projid = "";
my $projname = "";
my $userkey = "";
my $user = "";
my $wdir = "";
#my $keyfile = "";

# Just intermediate storage used by the parser.
my %commit_list = (); # branch -> { date -> message }

# The following are output from the parser
my %commit_order = (); # { branch -> [date, message] }
my @branch_order = (); # [ branch ]
my %branch_arcs = ();  # { branch -> parent_branch }
my %tags = ();         # { tag -> { file -> version } }
my %versions = ();     # { file -> { version -> [date, branch] }

my %tag2commit = ();   # { tag -> [date, branch] }

sub subtract_second {
    my ($date) = @_;
    
    if ($date =~ /^(\d+)\/(\d+)\/(\d+) (\d+):(\d+):(\d+) /) {
	my ($D_y,$D_m,$D_d, $Dh,$Dm,$Ds) = 
	    Add_Delta_YMDHMS($1,$2,$3, $4,$5,$6, 0, 0, 0, 0, 0, -1);
	
	return sprintf ("%04d/%02d/%02d %02d:%02d:%02d", $D_y, $D_m, $D_d, $Dh, $Dm, $Ds);
    } else {
	print "Error parsing date: $date\n";
	exit(1);
    }
}

sub transfer_pastwatch_dirs {
    my ($from, $to) = @_;
    @dirs0 = &existing_dirs($from);
    @dirs1 = &existing_dirs($to);

    foreach $d (@dirs0) {
	if ($d =~ /^(.*\/)Pastwatch/ && scalar(grep /^$1$/, @dirs1)==0) {
	    print "moving $from/$d to $to/$1\n";
	    execute("mkdir -p $to/$1; mv $from/$d $to/$1");
	}
    }
}

sub calculate_added_files {
    my ($old, $new) = @_;
    my @added;

    @files0 = &existing_files($old);
    @files1 = &existing_files($new);

    foreach $newfile (@files1) {
	if (scalar (grep /^$newfile$/, @files0) == 0) {
	    push @added, $newfile;
	    print "Commit added $newfile\n";
	}
    }
    return @added;
}
sub calculate_removed_files {
    my ($old, $new) = @_;
    my @removed;

    @files0 = &existing_files($old);
    @files1 = &existing_files($new);

    foreach $oldfile (@files0) {
	if (scalar (grep /^$oldfile$/, @files1) == 0) {
	    push @removed, $oldfile;
	    print "Commit removed $oldfile\n";
	}
    }
    return @removed;
}
sub existing_files {
    my ($dir) = @_;
    my $files = `cd $dir; find . | grep -v /Pastwatch`;
    return split /\n/, $files;
}

sub existing_dirs {
    my ($dir) = @_;
    my $dirs = `cd $dir; find . -type d`;
    return split /\n/, $dirs;
}

sub execute {
    my ($command) = @_;
    print "Executing: $command\n";

    my $r = system($command);
    if ($r != 0) {
	print "problem running: $r\n";
	exit($r);
    }
}

sub sort_dates {
    my ($branch) = @_;
    my %h = %{$commit_list{$branch}};
    my $x;
    my @a = sort {str2time("$a") <=> str2time("$b")} keys %h;
    foreach $x (@a) {
	my $t = str2time($x);
    }
    return @a;
}

sub push_commit {
    my ($date, $message, $branch) = @_;
    #print "Commit [$date, $branch, $message]\n";

    if (!$commit_list{$branch}) { $commit_list{$branch} = {}; }
    if ( ${%{$commit_list{$branch}}}{$date} &&
	"${%{$commit_list{$branch}}}{$date}" ne $message ) {
	print "Warning: multiple commits at $date on branch $branch ";
	print "have different commit messages. The messages will be ";
	print "appended.\n";
	print "  1: ${%{$commit_list{$branch}}}{$date}";
	print "  2: $message";

	my $s = "${%{$commit_list{$branch}}}{$date}";
	${%{$commit_list{$branch}}}{$date} = "$s $message"; 
    } else {
        ${%{$commit_list{$branch}}}{$date} = $message;
    }
}

sub escape_quotes {
    my ($s) = @_;
    
    $s =~ s/'/\\'/g;
    $s =~ s/"/\\"/g;
    return $s;
}
sub is_branch {
    my ($v) = @_;
    my @a = split /\./, $v;
    return  ((scalar(@a) % 2) != 0);
}
sub trim_one {
    my ($v) = @_;
    return substr($v, 0, rindex($v, "."));
}

sub de_vendor {
    my ($v) = @_;
    if ($v =~ /^1\.1\.(1\.1.*)/) { return $1; }
    return $v;
}

sub is_parent {
    my ($b1, $b2) = @_;
    # returns 1 if $b1 is a predecessor of $b2, whether 
    # directly or indirectly.  Assumes that %branch_arcs 
    # is already set.
    if ($b1 eq "init") { return 1; }

    my $b = $b2;
    while ($b ne "init") {
	if ($b eq $b1) { return 1; }
	$b = $branch_arcs{$b};
    }
    return 0;
}

sub parse_logs {
    my ($wd) = @_;
    execute ("cd $wd; cvs log > ../../cvslog");
    open (LOGFILE, "<cvslog") or die "can't open cvs log file";

    my ($workingfile, $headversion, $rev, $version);
    my @commit_versions = ();
    my @commit_dates = ();
    my @commit_messages = ();
    my %symbols = ("init" => "1");
    my %branches = ("init" => "1");
    my $parsing_symbols = 0;
    my ($x, $y, $i);
    my @arcs = ();
    my $reading_message = 0;
    my $message = "";
    my %version2branch = ();

    # Parse Log Output
    while (<LOGFILE>) {
	if ($parsing_symbols && /^\s+(\S+): (\S+)$/) {
	    my $b = $1;
	    $x = $2;
	    $x =~ s/\.0\./\./g;
	    $x = &de_vendor($x);
	    $symbols{$b} = $x;
	} elsif (/^revision (\S+)$/) {
	    if ($1 eq "1.1.1.1") {
		$rev = $1;
	    } else {
		$rev = &de_vendor($1);
	    }
	} elsif (/^date: (\S+ \S+);/) {
	    $x = $1;
	    $x =~ s/\.0\./\./g;
	    push @commit_versions, $rev;
	    push @commit_dates, $x;
	    #$versions{$workingfile}{$rev} = [$x];
	    $reading_message = 1;
	} elsif (/^----------------------------$/) {
	    if ($reading_message) {
		push @commit_messages, $message;
	    }
	    $reading_message = 0;
	    $message = "";
	} elsif (/^branches:\s+(\S+.*)$/) {
	    my @a = split /;\s*/, $1;
	    foreach $x (@a) { }
	} elsif (/^Working file: (\S+)$/) {
	    $workingfile = $1;
	} elsif (/^head: (\S+)$/) {
	    $headversion = $1;
	} elsif (/^symbolic names:$/) {
	    $parsing_symbols = 1;
	} elsif (/^keyword substitution:/) {
	    $parsing_symbols = 0;
	} elsif (/^=============================================================================$/) {
	    # done with log
	    if ($reading_message) {
		push @commit_messages, $message;
	    }
	    $reading_message = 0;

	    foreach $x (keys %symbols) {
		if (&is_branch($symbols{$x})) { 
		    $branches{$x} = 1;
		    $version2branch{$symbols{$x}} = $x;
		} else {
		    # It's a symbolic tag
		    if (!$tags{$x}) {$tags{$x} = {};}
		    $tags{$x}{$workingfile} = $symbols{$x};
		}
	    }
	    for ($i=0; $i<scalar(@commit_versions); $i++) {
		my $branch = $version2branch{&trim_one($commit_versions[$i])};
		#print "pushing commit ($workingfile) $commit_dates[$i], {$branch}\n";
		if (!$branch || $branch eq "") {
		    print "Warning: $workingfile revision $commit_versions[$i] ";
		    print "has no branch name. The branch name may have been removed.  ";
		    print "Ignoring the revision.\n";
		} else {
		    &push_commit("$commit_dates[$i] +000", 
				 &escape_quotes($commit_messages[$i]), $branch);
		
		    $versions{$workingfile}{$commit_versions[$i]} = 
			["$commit_dates[$i] +000", $branch];
		}
	    }

	    # Analyze Branch Structure for this file
	    # arcs uses the branch names to discover dependance
	    # XXX I should change this so that we only look at
	    #     descendants that actually have a version.  Not
	    #     just an entry in the symbol list
	    foreach $x (keys %branches) {
		foreach $y (keys %branches) {
		    if (&trim_one(&trim_one($symbols{$x})) eq $symbols{$y}) {
			push @arcs, [$y, $x];
 		    }
		}
	    }

	    @commit_versions = ();
	    @commit_dates = ();
	    @commit_messages = ();
	    %symbols = ("init" => "1");
	    %branches = ("init" => "1");
	    %version2branch = ();
	    $workingfile = "";
	    $headversion = "";
	    $rev = "";
	    $message = "";
	} elsif ($reading_message) {
	    $message = "$message$_";
	}
    }

    # We can reconstruct the branch structure now.
    my %final_arcs = ();
    my @remaining_arcs = @arcs;
    my %depths = ("init" => 0);
    my $current_depth = 0;

    while (scalar (@remaining_arcs) > 0) {
	my @t = ();

	# Find nodes at current depth
	my @at_depth = ();
	foreach $x (keys %depths) {
	    if ($depths{$x} == $current_depth) {
		push @at_depth, $x;
	    }
	}

	# Foreach node in current_depth, set pointer in final_arcs
	foreach $x (@at_depth) {
	    foreach $y (@remaining_arcs) {
		if ($x eq "${@{$y}}[0]") {
		    $final_arcs{ "${@{$y}}[1]" } = $x;
		    $depths{"${@{$y}}[1]"} = $current_depth + 1;
		} else {
		    push @t, $y;
		}
	    }
	}
	$current_depth++;
	@remaining_arcs = @t;
    }

    # Transfer final parser output
    %branch_arcs = %final_arcs;

    foreach $branch (keys %commit_list) {
	my @commits = &sort_dates($branch);
	$commit_order{$branch} = [];
	foreach $y (@commits) {
	    push @{$commit_order{$branch}}, [$y, "${%{$commit_list{$branch}}}{$y}"];
	}
    }

    # Condense the commits.  Currently, the commits are one per file per date.
    # We will change that to one per date.
    for($x=0; $x<=$current_depth; $x++) {
	foreach $y (keys %depths) {
	    if ($x == $depths{$y}) {
		push @branch_order, $y;
	    }
	}
    }
}

sub print_parse_output {
    my ($branch, $commit_pair, $time, $message, $tag, $file, $i);

    print "PARSER OUTPUT\n";

    foreach $branch (@branch_order) {
	print "Commit branches in order: $branch\n";
    }
    foreach $branch (keys %branch_arcs) {
	print "Branches $branch -> $branch_arcs{$branch}\n";
    }
    $i = 0;
    foreach $branch (keys %commit_order) {
	foreach $commit_pair (@{$commit_order{$branch}}) {
	    $time = @{$commit_pair}[0];
	    $message = @{$commit_pair}[1];
	    print "Commit $i ($branch) $time\n";
	    $i++;
	}
    }
    foreach $tag (sort keys %tags) {
	foreach $file (sort keys %{$tags{$tag}}) {
	    my $v = $tags{$tag}{$file};
	    if (!$v) {$v = "none";}
	    print "Tag: $tag - $file $v\n";
	}
    }
    foreach $file (sort keys %versions) {
	foreach $version (sort keys %{$versions{$file}}) {
	    my $d = $versions{$file}{$version}[0];
	    print "Version: $file - $version -> $d\n";
	}
    }
}

sub map_tags_to_versions {
    my ($tag, $file, $ver, $date, $x);
    
    foreach $tag (sort keys %tags) {
	my @descendants = (); # [ version, branch ]
	my %start_times = ();  # { file -> date }
	my %end_times = ();    # { file -> date }
	my $branch = "";
	foreach $file (sort keys %{$tags{$tag}}) {
	    $ver = $tags{$tag}{$file};
	    $date = $versions{$file}{$ver}[0];
	    if ($branch eq "") { $branch = $versions{$file}{$ver}[1]; }

	    # If they have different branches, and one branch is not a parent
	    # of the other, then complain.
	    # XXX later, we should check if the file was altered in the parent branch
	    # at a time earlier than the tag point.

	    if ($branch ne $versions{$file}{$ver}[1]) {
		if (&is_parent($branch, $versions{$file}{$ver}[1])) {
		    $branch = $versions{$file}{$ver}[1];
		} elsif (&is_parent($versions{$file}{$ver}[1], $branch)) {
		} else {
		    print "Warning: tag $tag mismatched ($file). ";
		    print "$branch != $versions{$file}{$ver}[1]\n";
		}
	    }
	    print "Reading Tag: $tag - $file $ver ($branch) - $date\n";
	    $start_times{$file} = $date;

	    foreach $x (&find_descendants($file, $ver)) {
		print "  Descendants: $x $versions{$file}{$x}[0] $versions{$file}{$x}[1] \n";
		push @descendants, [$versions{$file}{$x}[0], $versions{$file}{$x}[1]];
	    }
	}
	# find latest start time
	# find earliest end time
	# repository is stationary inbetween start and end.
	my $start = &latest_date(%start_times);
	my $end   = &earliest_date($branch, @descendants);
	if ($end eq "") { $end = $start; }

	print "Tag: $tag is on branch $branch from $start to $end\n";

	if (str2time($start) > str2time($end)) {
	    print "Warning: Tag $tag does not specify one coherent time period.  ";
	    print "Changing it to mark $start\n";
	    $end = $start;
	}
	
	my ($c, $i);
	my $tagged = -1;
	my @commits = @{$commit_order{$branch}};
	for ($i=0; $i<scalar(@commits); $i++) {
	    #print "  $i $commits[$i][0]\n";
	    if ( str2time($start) < str2time($commits[$i][0]) ) {
		$tagged = $i-1;
		last;
	    }
	    if ( str2time($start) == str2time($commits[$i][0]) ) {
		$tagged = $i;
		last;
	    }
	}
	if ($tagged < 0 && $i >= scalar(@commits)) { $tagged = $#commits; }
	if ($tagged < 0) { 
	    print "Warning: Could not find commit that maches Tag $tag from ";
	    print "$start to $end\n";
	    next;
	}

	# Mark the tagged version
	print "  Tagged $commits[$tagged][0]\n";
	$tag2commit{$tag} = [$commits[$tagged][0], $branch];
    }
}

sub find_descendants {
    my ($file, $ver) = @_;
    my ($v, $i, $match);
    my @versions = keys %{$versions{$file}};
    my @matches = ();

    my @a = split /\./, $ver;
    foreach $v (@versions) {
	#print "  comparing $ver to $v\n";
	my @b = split /\./, $v;
	$match = 1;

	if (scalar(@b) == scalar(@a)) {
	    for($i=0; $i<scalar(@a)-1; $i++) {
		if ($a[$i] != $b[$i]) {$match = 0;}
	    }
	    if ($match && $a[$#a]+1 != $b[$#b]) { $match = 0; }
	} elsif (scalar(@b) == scalar(@a)+2) { 
	    for($i=0; $i<scalar(@a); $i++) {
		if ($a[$i] != $b[$i]) {$match = 0;}
	    }
	    # Anything .1.1 is a branch, and did not change the file content.
	    # Anything .1.2 is a descendant.
	    if ($b[$#b] ne "2") { $match = 0; } 
	} else {
	    $match = 0;
	}
	if ($match) { push @matches, $v; }
    }
    return @matches;
}

sub earliest_date {
    my ($branch, @v) = @_;
    # Takes a list like this [ [date, branch] ]
    # and returns the smallest date who's branch == $branch.
    my $d;
    my $smallest_d = "";
    my $i;
    
    for($i=0; $i<scalar(@v); $i++) {
	if ($smallest_d eq "" || (str2time($v[$i][0]) < str2time($smallest_d))) { 
	    if ($branch eq $v[$i][1]) {
		$smallest_d = $v[$i][0]; 
	    }
	}
    }
    return $smallest_d;
}

sub latest_date {
    my %h = @_;
    # takes a hash like this { key -> date }
    # and returns the largest date;
    my $k;
    my $largest_k;
    foreach $k (keys %h) {
	if (!$largest_k || (str2time($h{$k}) > str2time($h{$largest_k}))) { 
	    $largest_k = $k; 
	}
    }
    return $h{$largest_k};
}

sub find_branchpoint {
    # Find & return the newest version that is older than $branchtime
    my ($parent, $branchtime, $dir) = @_;
    my $output = `cd $dir; pastwatch -b $parent -p $projid history`;
    my @lines = split /\n/, $output;
    my $version = "";
    my $time;
    my $targettime = str2time($branchtime);

    foreach $line (@lines) {
	if ($line =~ /^Version: (\S+) /) {
	    $version = $1;
	} elsif ($line =~ /^date: (.+)$/) {
	    $time = str2time("$1");
	    print " $version time, target: $1, $branchtime ($time, $targettime)\n";
	    if ($time <= $targettime) {return $version;}
	    $version = "";
	}
    }

    if ($version eq "" || ($time ne "" && $time > $targettime)) {
	print "Could not find branch point at $branchtime on branch $parent.\n";
	exit(1);
    }
    return $version;
}

sub init_trunk {
    my $date    = @{@{$commit_order{"init"}}[0]}[0];
    my $message = @{@{$commit_order{"init"}}[0]}[1];

    execute("cd $wdir; cvs -d $cvsroot export -D \"$date\" -d x $cvs_module");    
    print "Importing project\n";
    execute("cd $wdir/x; pastwatch -p $projname import -S \"$date\" -m \"$message\" $past_module");
    execute("rm -r $wdir/x");
}

sub commit_branch {
    my ($mybranch, $parent) = @_;

    my $branch = "";
    my $removeme;
    my $addme;
    my @removed_files;
    my @added_files;
    
    my $p_branch_parent = "";
    my $p_branch_child  = "";
    my $c_branch_parent = "";
    my $c_branch_child  = "";

    if (!$parent || $parent eq "") {
	$parent = "";
    } elsif ($parent eq "init") {
	$p_branch_parent = "-b $parent";
	$c_branch_parent = "";
    } else {
	$p_branch_parent = "-b $user:$parent";
	$c_branch_parent = "-r $parent";
    }
    
    if ($mybranch eq "init") {
	$p_branch_child = "-b $mybranch";
	$c_branch_child = "";
    } else {
	$p_branch_child = "-b $user:$mybranch";
	$c_branch_child = "-r $mybranch";
    }
    
    my @commits = @{$commit_order{$mybranch}};

    print "Checking out branch [$mybranch] of project parent is $parent\n";
    execute("cd $wdir; pastwatch $p_branch_parent -p $projid checkout -d a $past_module");

    # Move to the parent branch if necessary.
    if ($mybranch ne "init") { 
	my $p = ($parent eq "init") ? "init" : "$user:$parent";
	$branchpoint = &find_branchpoint($p, @{@{$commit_order{$mybranch}}[0]}[0], "$wdir/a"); 
	print "Branchpoint is: $branchpoint\n";

	my $time_prior = &subtract_second(@{@{$commit_order{$mybranch}}[0]}[0]);
	execute("cd $wdir/a; pastwatch $p_branch_parent -p $projid branch -f $branchpoint -S \"$time_prior\" -B -b $mybranch -m \"Starting branch $mybranch\"");
	execute("rm -r $wdir/a;");
	execute("cd $wdir; pastwatch -p $projid $p_branch_child checkout -d a $past_module");
    }

    for ($i=1; $i< scalar(@commits); $i++) {
	print "Commiting patch $i at $commits[$i][0] from branch $mybranch\n";

	execute("cd $wdir; cvs -d $cvsroot export $c_branch_child -D \"$commits[$i][0]\" -d b $cvs_module");    
	@removed_files = reverse sort &calculate_removed_files("$wdir/a", "$wdir/b");
	@added_files = sort &calculate_added_files("$wdir/a", "$wdir/b");

	foreach $removeme (reverse sort @removed_files) {
	    execute("cd $wdir/a; rm -rf $removeme");
	    execute("cd $wdir/a; pastwatch $p_branch_child -p $projid remove $removeme");
	    #XXX this is temporary during debug.
	    execute("cd $wdir/a; pastwatch $p_branch_child -p $projid commit -S \"$commits[$i][0]\" -m bla");
	}

	&transfer_pastwatch_dirs("$wdir/a", "$wdir/b");
	execute("rm -r $wdir/a");
	execute("mv $wdir/b $wdir/a");

	foreach $addme (@added_files) {
	    execute("cd $wdir/a; pastwatch $p_branch_child -p $projid add -S \"$commits[$i][0]\" $addme");
 	}	    

	my $message = $commits[$i][1];
	execute("cd $wdir/a; pastwatch $p_branch_child -p $projid commit -S \"$commits[$i][0]\" -m \"$message\"");
    }
    execute("rm -r $wdir/a");
}

sub commit_tags {
    my $tag;

    print "Checking out branch init of project for tagging\n";
    execute("cd $wdir; pastwatch -b init -p $projid checkout -d tagme $past_module");

    foreach $tag (keys %tag2commit) {
	my $date = $tag2commit{$tag}[0];
	my $branch = $tag2commit{$tag}[1];

	if ($branch ne "init") {
	    $branch = "$user:$branch";
	}
	print "Tagging branch $branch with $tag at $date\n";
	execute("cd $wdir/tagme;  pastwatch -b $branch -p $projid rtag -D \"$date\" $tag");
    }
}

sub create_proj_user {
    my ($username, $projname, $timestamp) = @_;
    my $output = `pastctl create $projname`;
    my @lines = split /\n/, $output;
    foreach $l (@lines) {
	if ($l =~ /new project ID: (\S+)$/) {
	    $projid = $1;
	}
    }
    if ($projid eq "") {
	print "Error creating proj\n";
	exit(1);
    }

    my $userid = "";
    $output = `pastctl join $username $projname`;
    @lines = split /\n/, $output;
    foreach $l (@lines) {
	if ($l =~ /^give (\S+) and/) {
	    $userid = $1;
	}
    }
    if ($userid eq "") {
	print "Error joining project\n";
	print "$output";
	exit(1);
    }

    $output = `pastctl addmember -S \"$timestamp\" -k $userid $username $projname`;

    @lines = split /\n/, $output;
    foreach $l (@lines) {
	if ($l =~ /new member: \S+, (\S+)$/) {
	    $userkey = $1;
	}
    }
    if ($userkey eq "") {
	print "Error adding member\n";
	print "$output";
	exit(1);
    }
}

sub usage {
    print "cvs2past.pl cvsroot cvsmodule pastmodule projname username\n";
    exit(1);
}

sub main {
    if (scalar(@ARGV) != 5) { &usage(); };

    $cvsroot     = shift @ARGV;
    $cvs_module  = shift @ARGV;
    $past_module = shift @ARGV;
    $projname    = shift @ARGV;
    $user        = shift @ARGV;

    $wdir = `mktemp -d cvs2pastwd.XXXXXX`;
    chomp $wdir;

    execute ("cd $wdir; cvs -d $cvsroot checkout -d x $cvs_module");
    &parse_logs("$wdir/x");
    &print_parse_output();
    &map_tags_to_versions();

    execute ("rm -r $wdir/x");

    &create_proj_user($user, $projname, @{@{$commit_order{"init"}}[0]}[0]);
    #$keyfile = "$pastwatch_root/$projname/key.$userkey";

    # Check in the commits
    &init_trunk();
    foreach $branch (@branch_order) {
	if ($branch ne "") {
	    &commit_branch($branch, $branch_arcs{$branch});
	}
    }

    # Add tags to the pastwatch repository
    &commit_tags();
    
    execute ("rm -r $wdir");
}

&main();
